#!/usr/bin/python3
# -*- coding: utf-8 -*-

import xmlutils
import geometry
import numpy as np
import argparse
import os
import sys
import subprocess
import util
import matplotlib.pyplot as plt
import copy
import tqdm
import matplotlib
import h5py
from swfileIO.ml_abn import ml_abn
import time
#from ML_FF_aux.mlip_lammps_interface import mlip_lammps_interface 

kJmol2eVatom = 0.0103642688
kcal2kJ = 4.186798
factor = kJmol2eVatom*kcal2kJ     # kcal/mol -> eV/atm
factor = 1.0

def get_angle(v, w): return np.arccos(v.dot(w)/(np.linalg.norm(v)*np.linalg.norm(w)))


def run_vasp_disp(fname,geom,disp_num=0,serial=False,write_only=False):
    
    disp_dir_name = "disp-%05d"%disp_num
    
    perform_calc = False
    if os.path.exists(disp_dir_name+"/vasprun.xml"):
        try:
            data = xmlutils.read_vasprun_xml(disp_dir_name+"/vasprun.xml")
            print("found and used existing displacement calculation for "+disp_dir_name)
        except:
            perform_calc = True
    else:
        perform_calc = True
    if perform_calc:
        util.mkdir(disp_dir_name)
        os.system("cp POTCAR INCAR "+disp_dir_name)
        os.chdir(disp_dir_name)
        os.system("ln -s "+fname+" ML_FF")
        geom.write_file("POSCAR",filetype="POSCAR")
        if not write_only:
            if not serial:
                os.system("mpirun vasp_std > vasp.out")
            else:
                os.system("vasp_std > vasp.out")
            data = xmlutils.read_vasprun_xml("vasprun.xml")
        os.chdir("..")
    if not write_only:
        forces = data["forces"]
        if "final_energy" in data.keys():
            energy = data["final_energy"]
        elif "energies" in data.keys():
            if len(data["energies"]) > 0:
                energy = data["energies"][-1]
            else:
                energy = 0
        else: 
            energy = 0
        if "stress" in data.keys():
            if len(data["stress"])!=3:
                stress = np.zeros((3,3))
            else:
                stress = data["stress"]
        else:
            stress = np.zeros((3,3))
        return energy, forces, stress
    else:
        return 0, np.zeros(np.shape(geom.coors)), np.zeros((3,3))

def run_mlp_disp(fname,geom,disp_num=0,serial=False):
    disp_dir_name = "disp-%05d"%disp_num
    util.mkdir(disp_dir_name)
    mobj = ml_abn(filetype="manual")
    mobj.from_geos([geom])
    os.chdir(disp_dir_name)
    #os.system("ln -s "+fname+" pot.mtp")
    mobj.write_MLIP_cfg("geom.cfg")
    if not serial:
        os.system("mpirun mlp calc-efs "+fname+" geom.cfg geom_efs.cfg")
    else:
        os.system("mlp calc-efs "+fname+" geom.cfg geom_efs.cfg")
    out_cfg = ml_abn(fname="geom_efs.cfg",filetype="cfg")
    forces = np.array(out_cfg.forces[0])
    energy = out_cfg.energies[0]
    stress = out_cfg.stress[0]
    os.chdir("..")
    return energy, forces, stress

def run_lammps_mlp_disp(mli,geom):
    geom.convert_cell_to_lammps() # this actually also rotates the reference forces, which makes the comparison not nonsensical
    mli.update_coors(geom.cell,geom.coors)
    energy = mli.calc_energy()
    forces = mli.get_force()
    stress = mli.get_stress_tensor()
    return energy, forces, stress

def write_2_column(fname,col1,col2):
    # simple script to write a 2 column file
    ofp = open(fname,"w")
    ofp.write("# %14s %14s\n" % ("REF","FF"))
    for cid,cval in enumerate(col1):
        ofp.write("  %14.8f %14.8f\n" % (cval,col2[cid]))
    ofp.close()
    return

def compare_forces_FF(fname, fffact=factor, direclvl=3, verbose=False, dispref="POSCAR", atom_plots=False, hdf5=False, code="vasp",
                      atom_template=None, from_cfg=None, serial=False, write_only=False):
  pltstyle = "~/ez-grp_style.mplstyle"
  plt.style.use(pltstyle)

  ml_ff_name = os.path.abspath(fname[0])
  #ref_direc_list = [i for i in os.listdir(fname[1]) if os.path.isdir(fname[1]+"/"+i)]  

  ### READ FORCES BLOCK START ###
  
  catenergies = {}
  catdisps = {}
  catrefforces = {}
  catffforces = {}
  catrefstresses = {}
  catffstresses = {}
  catrefdirs = {}
  coordiffs = {}
  cat_partial_energies = {}

  #bar = Bar('calculating POSCARS', max=len(ref_direc_list))
  rmsds = []
  rms_total = []
  stress_rmsds = []
  stress_rms_total = []
  maxsqdiffs = []
  dircats = []
  
  energy_file_names = []
  ref_energies = None
  # when reading from cfg and using VASP ML FF as a calculator - need to adjust elements for written POSCARs
  if from_cfg is not None:
      cfg_obj = ml_abn(".",filetype="none")
      cfg_obj.read_MLIP_cfg(from_cfg)
      geom_list = cfg_obj.get_geos(forcename="ref_forces",stress_format="matrix")
      ref_energies = cfg_obj.energies
      dircat = "cfg"
      dircats = [dircat] * len(geom_list)
      if dircat not in catenergies.keys():
          catenergies[dircat] = []
          catdisps[dircat] = []
          catrefforces[dircat] = []
          catffforces[dircat] = []
          catrefstresses[dircat] = []
          catffstresses[dircat] = []
          catrefdirs[dircat] = []
          coordiffs[dircat] = []
          cat_partial_energies[dircat] = {}
  else:
      p = subprocess.Popen(["find", fname[1], "-name", "vasprun.xml"], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      paths, err = p.communicate()
      ref_direc_list = paths.decode().split("\n")
      ref_direc_list.sort()
      ref_direc_list = ref_direc_list[1:]  # to avoid the empty line
      # read the energies here
    
      p2 = subprocess.Popen(["find", fname[1], "-name", "ENERGIES_*"], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      paths, err = p2.communicate()
      energy_file_list = paths.decode().split("\n")
      energy_file_list.sort()
    
      for entry in energy_file_list:
        energy_file_names.append(entry.split("/")[-1])
      
      geom_list = []
      
      print("reading forces from vasprun files...")
      
      for rdid in range(len(ref_direc_list)):
        ref_direc = ref_direc_list[rdid]
        # to assign the directory to a certain data set 
        # by default the datasets are built from the second to last directroy in the tree
        spdir = ref_direc.split("/")
        dircat = spdir[-direclvl]
        dircats.append(dircat)
        if dircat not in catenergies.keys():
          catenergies[dircat] = []
          catdisps[dircat] = []
          catrefforces[dircat] = []
          catffforces[dircat] = []
          catrefstresses[dircat] = []
          catffstresses[dircat] = []
          catrefdirs[dircat] = []
          coordiffs[dircat] = []
          cat_partial_energies[dircat] = {}
        catrefdirs[dircat].append(ref_direc)
        data = xmlutils.read_vasprun_xml(ref_direc)
        geom = geometry.geo(ref_direc.split("vasprun.xml")[0]+"POSCAR", filetype="POSCAR")
        geom.add_additional_data('ref_forces',data['forces'])
        geom.add_additional_data("stress",data['stress'])
        geom_list.append(geom)
          
      
      ### READ FORCES BLOCK END ###
  
  #print(ref_direc_list)

  #pd = pylmps.pylmps(molsys_name, screen=False, logfile="lmps.log")
  
  elapsed_time = 0


  def write_forces(fp, ref_force, fit_force, total_diff, sname):
    fp.write("System: %s  \n" % (sname))
    fp.write("# %8s %8s %12s   %12s   %12s\n" % ("atom_id","direc","fit_force","ref_force","abs_diff"))
    for vid, val in enumerate(fit_force):
      for direc in range(3):
        fp.write("  %8d %8d %12.8f   %12.8f   %12.8f\n" % (vid+1,direc+1,fit_force[vid,direc],ref_force[vid,direc],total_diff[vid,direc]))
        
      
  
  if atom_template is not None:
      mobj = ml_abn(".",filetype="none")
      mobj.read_MLIP_cfg(atom_template)
      atom_types = np.array(mobj.elems[0])
  else:
      atom_types = None
  testgeom = geometry.geo(dispref, filetype="POSCAR", atom_types=atom_types)
  #testgeom = geometry.geo("/home/sandro_w/data/VASP_data/IRMOF-10/rand_disps/POSCAR", filetype="POSCAR")
  basecoors = copy.copy(testgeom.coors)
  ellogs = []
  elems = testgeom.elems
  uelems = np.unique(testgeom.elems)
  elhdr = ""
  elrmsds = {}
  
  if code == "lammps_mlp":
      from ML_FF_aux.mlip_lammps_interface import mlip_lammps_interface 
      from mpi4py import MPI

      commw = MPI.COMM_WORLD
      rank = commw.Get_rank()
      
      
      testgeom.convert_cell_to_lammps()
      mli = mlip_lammps_interface()
      mli.load_mlip_potential(ml_ff_name,testgeom)
  else:
      rank = 0

  if verbose & (rank == 0):
    verbosefp = open("cart_forcecompare.dat", "w")
  
  for uel in uelems:
    ellogs.append(elems==uel)
    elhdr += "%12s " % (uel)
    elrmsds[uel] = []

  # stores values for a certain dataset (defined by the directory structure)
  
  if rank == 0:
    ofp = open("forcecompare.dat","w")
    ofp.write("#%6s %16s %12s %12s %s\n" % ("ID","NAME","RMSD","MAXSQDIFF",elhdr))

            

    
#  all_ric_forces = []
#  all_ref_ric_forces = []
#  n_intstrs = -1
  #coor_pointer = pd.lmp.extract_atoms()
  #pd.command("neigh_modify every 1")
#  newcell = None
  energy_sum = 0
  all_ffforces = []
  all_refforces = []
  all_refstresses = []
  all_ffstresses = []
  numatoms = {}
  for rdid in tqdm.tqdm(range(len(geom_list))):
    geom = geom_list[rdid]
    dircat = dircats[rdid]
    numatoms[dircat] = len(geom.coors)
    
    #dispnum = int(ref_direc.split("disp-")[-1].split("/")[0])
    #print(dispnum)
    
    
    
    coordiff = np.abs(basecoors-geom.coors)
    disp_dist = np.sum(np.linalg.norm(coordiff,axis=1))
    catdisps[dircat].append(disp_dist)
    coordiffs[dircat].append(coordiff)
    #geom.convert_cell_to_lammps()
    
 
    
    #pd = reload_pylmps(coors=geom.coors, cell=geom.cell)
    #pd.set_cell(geom.cell,cell_only=True)
    #pd.mol.set_cell(geom.cell)
    #pd.set_xyz(geom.coors)
    ctime = time.time()
    if code == "vasp":
        geom.elems = elems # need to adjust for VASP
        energy, aforces, stress = run_vasp_disp(ml_ff_name,geom,disp_num=rdid,serial=serial,write_only=write_only)
    elif code == "mlp":
        energy, aforces, stress = run_mlp_disp(ml_ff_name,geom,disp_num=rdid,serial=serial)
    elif code == "lammps_mlp":
        energy, aforces, stress = run_lammps_mlp_disp(mli,geom)
        # to comply with lammps, the stress needs to be in kbar (in metal units in lammps the stress is given in bar)
        stress /= 1000
    elapsed_time += time.time()-ctime
    
    energy_sum += energy
      
    #energy_dict = pd.get_energy_contribs()
    #for key in energy_dict.keys():
      #if key not in cat_partial_energies[dircat].keys():
        #cat_partial_energies[dircat][key] = []
      #cat_partial_energies[dircat][key].append(energy_dict[key])
    
    catenergies[dircat].append(energy)
    aforces = aforces * fffact
    refforces = geom.adata['ref_forces'] / factor * fffact
    #if ic_compare:
      #ric_forces = project_forces_to_ic(pd.mol, aforces, length = n_intstrs)
      #all_ric_forces.append(ric_forces)
      #ric_forces = project_forces_to_ic(pd.mol, refforces, length = n_intstrs)
      #all_ref_ric_forces.append(ric_forces)
      #if rdid == 0:
        #i_strs = get_interaction_strings(pd.mol,all_ref_ric_forces[0])
        #n_intstrs = len(i_strs)
      
    sqdiff = (refforces-aforces)**2
    rmsd = (np.sum(sqdiff)/len(sqdiff))**0.5
    maxsqdiff = np.max(sqdiff)
    maxsqdiffs.append(maxsqdiff)
    rmsds.append(rmsd)
    rms_total.append(np.sqrt(np.mean(refforces**2)))
    all_refforces.append(refforces)
    all_ffforces.append(aforces)
    
    if rank == 0:
        if len(refforces) != len(aforces):
            print("lenforce:",len(refforces),len(aforces))
        if np.shape(refforces) != np.shape(aforces):
            print("np.shapeforce:",np.shape(refforces),np.shape(aforces))
    catrefforces[dircat].append(refforces)
    catffforces[dircat].append(aforces)
    catrefstresses[dircat].append(geom.adata['stress'])
    catffstresses[dircat].append(stress)
    
    # for stresses
    if np.shape(geom.adata['stress']) == np.shape(stress):
        s_sqdiff = (geom.adata['stress']-stress)**2
        s_rmsd = (np.sum(s_sqdiff)/len(s_sqdiff))**0.5
        stress_rms_total.append(np.sqrt(np.mean(geom.adata['stress']**2)))
        stress_rmsds.append(s_rmsd)
        all_refstresses.append(geom.adata['stress'])
        all_ffstresses.append(stress)
    else:
        print("ERROR: inconsistent stress shape:",np.shape(stress),"compared to",np.shape(geom.adata['stress']))
  #  geom.write_file("POSCAR-%03d" % (rdid), filetype="POSCAR")
    
    uelstr = ""
    for uid,uel in enumerate(uelems):
      elsqdiff = sqdiff[ellogs[uid]]
      elrmsd = (np.sum(elsqdiff)/len(elsqdiff))**0.5
      uelstr += "%12.6f " % (elrmsd)
      elrmsds[uel].append(elrmsd)
    if rank == 0:
        if verbose:
            write_forces(verbosefp, refforces, aforces, abs(refforces-aforces), ref_direc)
        ofp.write(" %6d %16s %12.6f %12.6f %s\n" % (rdid,dircat,rmsd,maxsqdiff,uelstr))
        
    #print sqdiff
    #print np.sum(sqdiff)/len(sqdiff)
    #print (geom.adata['ref_forces'],aforces)
    
    #for fid,force in aforces:
    
  
  if verbose:
    verbosefp.close()

  rmsds = np.asarray(rmsds)
  maxsqdiffs = np.asarray(maxsqdiffs)

  etotrmsd = ""
  emaxrmsd = ""
  emeanrmsd = ""
  emedianrmsd = ""
  for uid,uel in enumerate(uelems):
    etotrmsd += "%12.6f " % (np.sum(elrmsds[uel]))
    emaxrmsd += "%12.6f " % (np.max(elrmsds[uel]))
    emeanrmsd += "%12.6f " % (np.mean(np.abs(elrmsds[uel])))
    emedianrmsd += "%12.6f " % (np.median(np.abs(elrmsds[uel])))
    
    
  if rank == 0:
    ofp.write("\n")
    ofp.write(" %20s %12.6f %s\n" % ("TOTAL RMSD:",np.sum(rmsds),etotrmsd))
    ofp.write(" %20s %12.6f %s\n" % ("MAX RMSD:",np.max(rmsds),emaxrmsd))
    ofp.write(" %20s %12.6f %s\n" % ("MEAN RMSD:",np.mean(np.abs(rmsds)),emeanrmsd))
    ofp.write(" %20s %12.6f %s\n" % ("MEDIAN RMSD:",np.median(np.abs(rmsds)),emedianrmsd))
    ofp.write(" %20s %12.6f\n" % ("MAX SQDIFF:",np.max(maxsqdiffs)))
    ofp.write(" %20s %12.6f\n" % ("MEAN(RMSD)/MEAN(RMS)",np.mean(rmsds)/np.mean(rms_total)))
    ofp.write(" %20s %12.6f\n" % ("MEAN(RMSD/RMS)",np.mean(rmsds/rms_total)))
    all_refforces = np.array(all_refforces)
    all_ffforces = np.array(all_ffforces)
    sqdiff = (all_refforces-all_ffforces)**2
    totrmsd = np.mean(sqdiff)**0.5
    totrms = np.sqrt(np.mean(all_refforces**2))
    totmsd = np.mean(sqdiff)
    totms = totrms**2
    ofp.write(" %20s %12.6f\n" % ("RMSD_ALL",totrmsd))
    ofp.write(" %20s %12.6f\n" % ("RMSD_ALL/RMS",totrmsd/totrms))
    ofp.write(" %20s %12.6f\n" % ("MSD_ALL",totmsd))
    ofp.write(" %20s %12.6f\n" % ("MSD_ALL/MS_ALL",totmsd/totms))
    
    ofp.close()
    
    
    with open("stresscompare.dat","w") as ofp2:
        ofp2.write("#%6s %16s %12s (kbar)\n" % ("ID","NAME","RMSD"))
        for eid,entry in enumerate(stress_rmsds):
            ofp2.write(" %6d %16s %12.6f\n" % (eid,dircats[eid],entry))
        ofp2.write(" %20s %12.6f\n" % ("TOTAL RMSD:",np.sum(stress_rmsds)))
        ofp2.write(" %20s %12.6f\n" % ("MAX RMSD:",np.max(stress_rmsds)))
        ofp2.write(" %20s %12.6f\n" % ("MEAN RMSD:",np.mean(np.abs(stress_rmsds))))
        ofp2.write(" %20s %12.6f\n" % ("MEDIAN RMSD:",np.median(np.abs(stress_rmsds))))
        ofp2.write(" %20s %12.6f\n" % ("MEAN(RMSD)/MEAN(RMS)",np.mean(stress_rmsds)/np.mean(stress_rms_total)))
        ofp2.write(" %20s %12.6f\n" % ("MEAN(RMSD/RMS)",np.mean(np.array(stress_rmsds)/np.array(stress_rms_total))))
        all_refstresses = np.array(all_refstresses)
        all_ffstresses = np.array(all_ffstresses)
        sqdiff = (all_refstresses-all_ffstresses)**2
        totrmsd = (np.mean(sqdiff))**0.5
        totrms = np.sqrt(np.mean(all_refstresses**2))
        totmsd = np.mean(sqdiff)
        totms = totrms**2
        ofp2.write(" %20s %12.6f\n" % ("RMSD_ALL",totrmsd))
        ofp2.write(" %20s %12.6f\n" % ("RMSD_ALL/RMS",totrmsd/totrms))
        ofp2.write(" %20s %12.6f\n" % ("MSD_ALL",totmsd))
        ofp2.write(" %20s %12.6f\n" % ("MSD_ALL/MS_ALL",totmsd/totms))
        ofp2.close()



  # here make category related evaluation
    dirname = "compare_forces_FF_plots"
    util.mkdir(dirname)
    all_energies = []
    all_energies_per_atom = []
    for key in catenergies.keys():
      xcoor = catdisps[key]
      #plt.plot(xcoor,catenergies[key])
      #xcoor = range(len(catenergies[key]))
      plt.scatter(xcoor,(np.array(catenergies[key])-min(catenergies[key]))*factor,label="FF")
      all_energies.append(catenergies[key])
      all_energies_per_atom.append(np.array(catenergies[key])/numatoms[key])
      # this is very specific to the VASP displacements Tomas calculated
      e_index = key.split("-")[-1]
      expected_energy_file_name = "ENERGIES_"+e_index+".dat"
      
      
      if expected_energy_file_name in energy_file_names:
        efind = energy_file_names.index(expected_energy_file_name)
        energy_path = energy_file_list[efind]
        efp = open(energy_path,"r")
        refenergies = []
        for line in efp:
          elements = line.split()
          if len(elements) > 0:
            refenergies.append(float(elements[0]))
        xcoor_ref = xcoor
        # need to resort the x coordinates if they are not sorted!
        #dirlist = catrefdirs[key]
        #indexlist = []
        #for el in dirlist:
          #dispindex = int(el.split("disp-")[-1].split("/")[0])
          #indexlist.append(dispindex)
        #xcoor_ref = np.array(xcoor)[indexlist]
        plt.scatter(xcoor_ref,np.array(refenergies)-min(refenergies),c="r",label="reference")
        plt.legend()
        
      plt.title(key)
      plt.xlabel("displacement distance / $\AA$")
      plt.ylabel("energy / eV")
      ax = plt.gca()
      ax.ticklabel_format(useOffset=False)
      plt.tight_layout()
      plt.savefig(dirname + "/energy_" + key+".png")
      plt.clf()
      
      
      for energy_key in cat_partial_energies[key].keys():
        energy_vals = np.array(cat_partial_energies[key][energy_key])
        plt.scatter(xcoor,energy_vals*factor,label=energy_key)
      plt.title(key)
      plt.legend()
      plt.xlabel("displacement distance / $\AA$")
      plt.ylabel("energy / eV")
      ax = plt.gca()
      ax.ticklabel_format(useOffset=False)
      plt.tight_layout()
      plt.savefig(dirname + "/energy_contribs_" + key+".png")
      plt.clf()
      
      
#      maxediff = max(catenergies[key])-min(catenergies[key])
      
      numforces = len(np.array(catrefforces[key])[0,:])*3
      colordist = np.array(([xcoor]*numforces)).T.ravel()
      allrefforces = np.array(catrefforces[key])
      allffforces = np.array(catffforces[key])
      
      reff = allrefforces.ravel()
      fff  = allffforces.ravel()
      plt.plot([min(reff),max(reff)],[min(reff),max(reff)],"k--")
      plt.scatter(reff,fff,c=colordist)
      cb = plt.colorbar()
      cb.set_label("displacement distance / $\AA$")
      plt.title(key)
      plt.xlabel("ref forces / eV/$\AA$")
      plt.ylabel("ff forces / eV/$\AA$")
      ax = plt.gca()
      ax.ticklabel_format(useOffset=False)
      plt.tight_layout()
      plt.savefig(dirname + "/forces_" + key+".png")
      plt.clf()
      
      if atom_plots:
        subdirname = dirname + "/atoms_" + key
        util.mkdir(subdirname)
        for aid, atomel in enumerate(testgeom.elems):
          atom_ref_forces = allrefforces[:,aid]
          atom_ff_forces  = allffforces[:,aid]
          allcoordiffs = np.array(coordiffs[key])
          dispvec = allcoordiffs[:,aid]
          atom_disp_dists = np.linalg.norm(dispvec, axis = 1)
          atom_ref_force_norms = np.linalg.norm(atom_ref_forces, axis = 1)
          atom_ff_force_norms = np.linalg.norm(atom_ff_forces, axis = 1)
          #angle = get_angle(atom_ref_forces[1],atom_ff_forces[1])
          #print(angle/2/np.pi*360)
          
          plt.scatter(atom_disp_dists,atom_ff_force_norms,c="b",label="ff")
          plt.scatter(atom_disp_dists,atom_ref_force_norms,c="r",label="ref")
          plt.xlabel("displacement distance / $\AA$")
          plt.ylabel("|force| / eV/$\AA$")
          plt.legend()
          
          plt.title("%s %5d " % (atomel,aid))
          
          plt.tight_layout()
          plt.savefig("%s/%s_%05d_.png" % (subdirname,atomel,aid))
          plt.clf()

#    minv = 0
#    maxv = 0

    
        
      
    #
    all_reff = None
    all_fff = None
    # plot all forces in a single plot
    for key in catrefforces.keys():
    ##  xcoor = catdisps[key]
    ##  maxediff = max(catenergies[key])-min(catenergies[key])
    ##  
    ##  numforces = len(np.array(catrefforces[key])[0,:])*3
    ##  colordist = np.array(([xcoor]*numforces)).T.ravel()
      if (np.shape(catrefforces[key])!=np.shape(catffforces[key])):
        print(np.shape(catrefforces[key]),np.shape(catffforces[key]))
      if all_reff is None:
          all_reff = np.array(catrefforces[key])
          all_fff = np.array(catffforces[key])
      else:
          all_reff = np.concatenate((all_reff,np.array(catrefforces[key])))
          all_fff = np.concatenate((all_fff,np.array(catffforces[key])))
    #  if minv > min(reff):
    #    minv = min(reff)
    #  if maxv < max(reff):
    #    maxv = max(reff)
    #  reff = allrefforces.ravel()
    #  fff  = allffforces.ravel()
    #  plt.scatter(reff,fff,c="b")#,c=colordist)
    #all_reff = np.array(all_reff)
    #all_fff = np.array(all_fff)
    plt.hist2d(all_reff.ravel(), all_fff.ravel(), cmap="plasma", bins=200, norm=matplotlib.colors.LogNorm())
    #plt.plot([minv,maxv],[minv,maxv],"k--")
    plt.plot([np.min(all_reff),np.max(all_reff)],[np.min(all_reff),np.max(all_reff)],"k--")
    cb = plt.colorbar()
    cb.set_label("amount")
    plt.xlabel("ref forces / eV/$\AA$")
    plt.ylabel("ff forces / eV/$\AA$")
    ax = plt.gca()
    ax.ticklabel_format(useOffset=False)
    plt.tight_layout()
    plt.savefig(dirname + "/forces_all_log.png")

    plt.clf()
    plt.hist2d(all_reff.ravel(), all_fff.ravel(), cmap="plasma", bins=200)
    #plt.plot([minv,maxv],[minv,maxv],"k--")
    cb = plt.colorbar()
    cb.set_label("amount")
    plt.xlabel("ref forces / eV/$\AA$")
    plt.ylabel("ff forces / eV/$\AA$")
    ax = plt.gca()
    ax.ticklabel_format(useOffset=False)
    plt.tight_layout()
    plt.savefig(dirname + "/forces_all.png")
    print("RMSE total:")
    total_rmse = util.rmse(all_reff.ravel(), all_fff.ravel())
    print(total_rmse)
    
    
    quantile_range=[0.05,0.95]
    
    # violin plot for absolute force differences
    plt.clf()
    parts = plt.violinplot(all_reff.ravel()-all_fff.ravel(),positions=[0],
                           showextrema=False,quantiles=quantile_range)
    write_2_column(dirname+"/all_force_vals.dat",all_reff.ravel(),all_fff.ravel())
    for pc in parts['bodies']:
        pc.set_facecolor('red')
        pc.set_edgecolor('black')
        pc.set_alpha(0.25)
    quant = parts["cquantiles"]
    quant.set_color("r")
    segs = quant.get_segments()
    quant_l = segs[0][0,1]
    quant_h = segs[1][0,1]
    print(quantile_range[0], " quantile: ",quant_l,"eV/A")
    print(quantile_range[1], " quantile: ",quant_h,"eV/A")
    plt.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=False)
    plt.ylabel("$\Delta F$ / eV/$\AA$")
    plt.xlabel("density")
    ax = plt.gca()
    ax.ticklabel_format(useOffset=False)
    plt.tight_layout()
    plt.savefig(dirname + "/forces_all_violin.png")
    
    # violin plot for absolute forces in the reference data set
    plt.clf()
    parts = plt.violinplot(all_reff.ravel(),positions=[0],
                           showextrema=False,quantiles=quantile_range)
    for pc in parts['bodies']:
        pc.set_facecolor('red')
        pc.set_edgecolor('black')
        pc.set_alpha(0.25)
    quant = parts["cquantiles"]
    quant.set_color("r")
    segs = quant.get_segments()
    rquant_l = segs[0][0,1]
    rquant_h = segs[1][0,1]
    print("reference forces: ", quantile_range[0], " quantile: ",rquant_l,"eV/A")
    print("reference forces: ", quantile_range[1], " quantile: ",rquant_h,"eV/A")
    print("force error ", quantile_range[0], ": ", quant_l/rquant_l*100,"%")
    print("force error ", quantile_range[1], ": ", quant_h/rquant_h*100,"%")
    print("force error difference: ", (quant_h-quant_l)/(rquant_h-rquant_l)*100,"%")
    plt.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=False)
    plt.ylabel("reference forces / eV/$\AA$")
    plt.xlabel("density")
    ax = plt.gca()
    ax.ticklabel_format(useOffset=False)
    plt.tight_layout()
    plt.savefig(dirname + "/forces_ref_violin.png")
    
    
    # plot all stresses in a single figure
    plt.clf()
    minval = 1e6
    maxval = 0
    for key in catrefstresses.keys():
        plt.scatter(catrefstresses[key],catffstresses[key],label=key)
        minval = np.min([np.min(catrefstresses[key]),minval])
        maxval = np.max([np.max(catrefstresses[key]),maxval])
    plt.plot([minval,maxval],[minval,maxval],"k--")
    plt.xlabel("ref stress / kbar")
    plt.ylabel("ff stress / kbar")
    plt.legend()
    plt.tight_layout()
    plt.savefig(dirname+"/stresses_all.png")
    
    plt.clf()
    parts = plt.violinplot(all_refstresses.ravel()-all_ffstresses.ravel(),positions=[0],
                           showextrema=False,quantiles=quantile_range)
    write_2_column(dirname+"/all_stress_vals.dat",all_refstresses.ravel(),all_ffstresses.ravel())
    for pc in parts['bodies']:
        pc.set_facecolor('blue')
        pc.set_edgecolor('black')
        pc.set_alpha(0.25)
    quant = parts["cquantiles"]
    quant.set_color("b")
    segs = quant.get_segments()
    quant_l = segs[0][0,1]
    quant_h = segs[1][0,1]
    print(quantile_range[0], " quantile: ",quant_l,"kbar")
    print(quantile_range[1], "quantile: ",quant_h,"kbar")
    plt.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=False)
    plt.ylabel("$\Delta \sigma$ / kbar")
    plt.xlabel("density")
    ax = plt.gca()
    ax.ticklabel_format(useOffset=False)
    plt.tight_layout()
    plt.savefig(dirname + "/stresses_all_violin.png")
    
    # violin plot for absolute stresses in the reference data set
    plt.clf()
    parts = plt.violinplot(all_refstresses.ravel(),positions=[0],
                           showextrema=False,quantiles=quantile_range)
    for pc in parts['bodies']:
        pc.set_facecolor('blue')
        pc.set_edgecolor('black')
        pc.set_alpha(0.25)
    quant = parts["cquantiles"]
    quant.set_color("b")
    segs = quant.get_segments()
    rquant_l = segs[0][0,1]
    rquant_h = segs[1][0,1]
    print("reference stresses: ", quantile_range[0], " quantile: ",rquant_l,"kbar")
    print("reference stresses: ", quantile_range[1], " quantile: ",rquant_h,"kbar")
    print("stress error ", quantile_range[0], ": ", quant_l/rquant_l*100,"%")
    print("stress error ", quantile_range[1], ": ", quant_h/rquant_h*100,"%")
    print("stress error difference: ", (quant_h-quant_l)/(rquant_h-rquant_l)*100,"%")
    plt.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=False)
    plt.ylabel("reference stresses / kbar")
    plt.xlabel("density")
    ax = plt.gca()
    ax.ticklabel_format(useOffset=False)
    plt.tight_layout()
    plt.savefig(dirname + "/stresses_ref_violin.png")
    
    
    if ref_energies is not None:
        print("energy RMSD:", util.rmse(ref_energies,np.array(all_energies).ravel()), " eV")
        # HACK for the reference number of atoms
        print("energy RMSD:", util.rmse(ref_energies/numatoms[list(numatoms.keys())[0]],np.array(all_energies_per_atom).ravel()), " eV/atom")
        print("energy MAD:", np.mean(np.abs(ref_energies/numatoms[list(numatoms.keys())[0]]-np.array(all_energies_per_atom).ravel())))
        
        write_2_column(dirname+"/all_energy_vals.dat",ref_energies,np.array(all_energies).ravel())
    
    
    if hdf5:
      ofh = h5py.File("forcecompare_data.hdf5", 'w')
      ofh.create_dataset("FF_forces", data=all_fff)
      ofh.create_dataset("ref_forces", data=all_reff)
      #if ic_compare:
        #ofh.create_dataset("FF_ric_forces", data=all_ric_forces)
        #ofh.create_dataset("ref_ric_forces", data=all_ref_ric_forces)
        ## these strings assume a consistent connectivity setup of all systems in
        ## the comparison
        ##i_strs = get_interaction_strings(pd.mol,all_ref_ric_forces[0])
        #ofh.create_dataset("interaction_strings", data=i_strs)
    return total_rmse, elapsed_time
  
      
if __name__ == "__main__":
  parser = argparse.ArgumentParser(description="compares forces from a dft reference with a vasp machine learned force field potential")



  parser.add_argument("-fffact", dest="fffact", type=float, default=1.0, 
                      help="factor for ff forces")
  parser.add_argument("-direclvl", dest="direclvl", type=int, default=3, 
                      help="defines the directory level which will be used to categorize the data set")
  parser.add_argument("-v", dest="verbose", action='store_true', default=False, 
                      help="verbose output of the computed forces")
  parser.add_argument("-serial", dest="serial", action='store_true', default=False, 
                      help="compute displacements in serial to save RAM")
  parser.add_argument("-dispref", dest="dispref", type=str, default="POSCAR", 
                      help="specify a reference file to compute the displacement distance")
  parser.add_argument("-from_cfg", dest="from_cfg", type=str, default=None, 
                      help="read the reference structures from a .cfg file instead")
  parser.add_argument("-write_only", dest="write_only", action="store_true", default=False, 
                      help="only write the displacements directories for VASP MLFF")
  parser.add_argument("-atom_plots", dest="atom_plots", action='store_true', default=False, 
                      help="plot displacement plots for each individual atom and its forces")

  parser.add_argument("-hdf5", dest="hdf5", action='store_true', default=False, 
                      help="store the forces in an hdf5 file forcecompare_data.hdf5 for further processing.")
  parser.add_argument("-code", dest="code", type=str, default="vasp", 
                      help="which code should be used to carry out the force predictions (vasp, mlp or lammps_mlp, default vasp)")
  parser.add_argument("-template", dest="atom_template", type=str, default=None, 
                      help="provide a template for the atom types used in a MLP type MTP in the form of a cfg file. The atoms have to be sorted in the same way as in the reference geometries. The masses will be taken from the actual elements")
  parser.add_argument("fname", nargs = 2, help="path to the ML FF potential file; 2 directory containing the DFT reference data")


  args = parser.parse_args()
  
  afp = open("p_cmnd.log","a")
  afp.write(" ".join(sys.argv)+"\n")
  afp.close()
  
  
  
  compare_forces_FF(**vars(args))
  
  

