#!/usr/bin/python3

# this script serves to compute a large set of single points with the MTP force field potential using lammps

import argparse
import geometry
import sys
import numpy as np
import tqdm
import copy
import util
from ML_FF_aux.mlip_lammps_interface import mlip_lammps_interface
import swfileIO.ml_abn as ml_abn

from mpi4py import MPI

commw = MPI.COMM_WORLD
rank = commw.Get_rank()
nprocs = commw.Get_size()


def write_forces(fp,forces,phpy=None):
    if phpy is not None:
        fp.write("\n")  # this would have to be added for phonopy
        fp.write("%d\n" % (phpy[0]))
        fp.write("   %18.10f %18.10f %18.10f\n" % (phpy[1][0],phpy[1][1],phpy[1][2]))
    else:
        fp.write("# \n# \n") # in phpy these two lines need to be filled TODO
    for force in forces:
        fp.write("%16.10f %16.10f %16.10f\n"%(force[0],force[1],force[2]))

        
def read_phpy_disps(filename):
    fp = open(filename)
    displacements = []
    atom_index = -1
    nprimatoms = 0
    read_uc = False
    for line in fp:
        elements = line.split()
        if len(elements) == 1:
            if elements[0] == "unit_cell:":
                read_uc = True
            elif elements[0] == "supercell:":
                read_uc = False
        if len(elements) >= 3:
            if elements[1] == "atom:":
                atom_index = int(elements[2])
            if (elements[1] == "symbol:") & (read_uc):
                nprimatoms += 1
        if (len(elements) == 5) & (atom_index != -1):
            dx = float(elements[1].split(",")[0])
            dy = float(elements[2].split(",")[0])
            dz = float(elements[3])
            displacements.append([atom_index,[dx,dy,dz]])
    return displacements, nprimatoms

def compute_disps(filelist,filetype="POSCAR",potfile="pot.mtp",FF="MLIP",fc_out=None,
                  phpy=None,num_disps=None, atom_template=None):
    
    mli = None
    
    if num_disps is not None:
        filelist = []
        if rank == 0:
            print("computing POSCAR-%05d to POSCAR-%05d"%(1,num_disps))
        for i in range(num_disps):
            if phpy is not None:
                filelist.append("POSCAR-%03d"%(i+1))
            else:
                filelist.append("POSCAR-%05d"%(i+1))
                
    if filetype == "POSCARS_all":
        fp_all = open(filelist[0],"r")
        first_line = fp_all.readline().split()
        num_structures = int(first_line[-1])
    else:
        num_structures = len(filelist)

            
    if phpy is not None:
        phpy_format = True
        disp_entries, nprimatoms = read_phpy_disps(phpy)
    else:
        phpy_format = False
        disp_entries = None
        
    if atom_template is not None:
        mobj = ml_abn.ml_abn(".",filetype="none")
        mobj.read_MLIP_cfg(atom_template)
        atom_types = np.array(mobj.elems[0])
    else:
        atom_types = None
        
    if rank == 0:
        if fc_out is not None:
            ofp = open(fc_out,"w")
        e_ofp = open("MLP_energies.dat", "w")
    for fid in tqdm.tqdm(range(num_structures)):
        
        if filetype == "POSCARS_all":
            disp_name = fp_all
        else:
            disp_name = filelist[fid]
            
        # this already reads the first structure in POSCARS_all mode which then has to be used for the mli initialization
        geom = geometry.geo(disp_name,filetype=filetype, atom_types=atom_types)
        startcell = copy.copy(geom.cell)
        geom.convert_cell_to_lammps()
        
        natoms = len(geom.coors)
        if mli is None:
#            testgeom = geometry.geo(disp_name, filetype=filetype, atom_types=atom_types)
#            testgeom.convert_cell_to_lammps()
            mli = mlip_lammps_interface()
            mli.load_mlip_potential(potfile,geom)
#            lmps = load_mlip_potential(potfile,testgeom)
            if (fc_out is not None) & phpy_format & (rank==0):
                ofp.write("%d\n"%natoms)
                ofp.write("%d\n"%len(filelist))
        mli.update_coors(geom.cell,geom.coors)
        forces = mli.calc_force()
        energy = mli.get_energy()
        if rank == 0:
            e_ofp.write("%6d %12.8f\n" % (fid, energy))
            # have to rotate forces back if the cell was rotated to comply with the lammps format
            if not np.allclose(startcell,geom.cell):
                forces = util.rotate_vec(forces,geom.cell,startcell)
                
            if fc_out is not None:
                if phpy_format:
                    phdisp = disp_entries[fid]
                else:
                    phdisp = None
                write_forces(ofp,forces,phpy=phdisp)
    if (fc_out is not None) & (rank == 0):
        print("\nforces written in "+fc_out)
        ofp.close()
        e_ofp.close()
    
    
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="computes displacements with mlip mtp potentials")


    parser.add_argument("-filetype", dest="filetype", type=str, default="POSCAR", help="filetype of the structures. Use type 'POSCARS_all' to read a sequence of files from a single file")
    parser.add_argument("-FF", dest="FF", type=str, default="MLIP", help="which type of FF should be applied")
    parser.add_argument("-potfile", dest="potfile", type=str, default="pot.mtp", help="name of the file that stores the FFP")
    parser.add_argument("-force_out", dest="fc_out", type=str, default="FORCE_SETS", help="name of the file in which the forces are written")
    parser.add_argument("-num_disps", dest="num_disps", type=int, default=None, help="optionally specify a number of displacements called POSCAR-* in case the file list becomes to large. This will ignore the main argument. This is meant for phono3py calculations, therefore the numbered POSCARs require 5 digits minimum.")
    parser.add_argument("-phpy", dest="phpy", type=str, default=None, help="format the force constants as phonopy does, have to specify the path to the phonopy_disp.yaml file")
    parser.add_argument("-template", dest="atom_template", type=str, default=None, 
                      help="provide a template for the atom types used in a MLP type MTP in the form of a cfg file. The atoms have to be sorted in the same way as in the reference geometries. The masses will be taken from the actual elements")
    parser.add_argument("filelist", nargs = "*", help="path to the structures")


    args = parser.parse_args()
    
    if rank==0:
        afp = open("p_cmnd.log","a")
        afp.write(" ".join(sys.argv)+"\n")
        afp.close()



    compute_disps(**vars(args))
    
    commw.Disconnect    
