#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
 This script serves to compute the phonon band structure from an mlip mtp
"""

import argparse
import sys
from ML_FF_aux.mlip_lammps_interface import mlip_lammps_interface
import geometry
import phonopy as ph
import numpy as np
import swfileIO.ml_abn as ml_abn
import util
import copy
from tqdm import tqdm

from mpi4py import MPI

commw = MPI.COMM_WORLD
rank = commw.Get_rank()
nprocs = commw.Get_size()

def compute_mlip_band(filename, filetype="POSCAR", potfile="pot.mtp", atom_template=None, 
                         disp=0.01, relax=False, relaxint=False, dim=[1,1,1], silent=False,
                         symprec=1e-5, txtFC=False, charges=None, charge_type="gauss/dsf",
                         escut=None, uelemsort=False):
    
    if len(dim) == 3:
        supercell = [[float(dim[0]),0.0,0.0],[0.0,float(dim[1]),0.0],[0.0,0.0,float(dim[2])]]
    elif len(dim) == 9:
        supercell = [[float(dim[0]),float(dim[1]),float(dim[2])],
                      [float(dim[3]),float(dim[4]),float(dim[5])],
                      [float(dim[6]),float(dim[7]),float(dim[8])]]
    else:
        print ("ERROR: invalid super cell definition")
        return None
    supercell = np.asarray(supercell)
    
    # load lammps interface
    if atom_template is not None:
        mobj = ml_abn.ml_abn(".",filetype="none")
        mobj.read_MLIP_cfg(atom_template)
        atom_types = np.array(mobj.elems[0])
    else:
        atom_types = None
    testgeom = geometry.geo(filename, filetype=filetype, atom_types=atom_types)
    init_geom = copy.deepcopy(testgeom)
    testgeom.convert_cell_to_lammps()
    rotated_cell = copy.copy(testgeom.cell)
    mli = mlip_lammps_interface(escut=escut)
    
    if charges is None:
        mli.load_mlip_potential(potfile, testgeom, uelemsort=uelemsort)
    else:
        mli.load_hybrid_mlip_charge_potential(potfile, testgeom, charges, charge_style=charge_type)
    
    # relax the box
    if relax:
        if (not silent) & (rank == 0):
            print("performing lattice relaxation...")
        mli.relax_box()
    elif relaxint:
        if (not silent) & (rank == 0):
            print("performing relaxation of internal coordinates...")
        mli.relax_int()
        
    if relax or relaxint:
        testgeom.coors = mli.get_xyz()
        testgeom.cell = mli.get_cell()
        init_geom.coors = util.rotate_vec(testgeom.coors, rotated_cell, init_geom.cell)
        init_geom.cell  = util.rotate_vec(testgeom.cell, rotated_cell, init_geom.cell)
        if rank==0:
            init_geom.write_file("POSCAR_relaxed",filetype="POSCAR")
    
    # generate phonopy displacements
    unitcell = ph.structure.atoms.PhonopyAtoms(symbols=list(init_geom.elems), cell=init_geom.cell, positions=init_geom.coors)
    if (not silent) & (rank == 0):
        print ("generating displacements for a \n %f %f %f\n %f %f %f\n %f %f %f \nsupercell" %
               (supercell[0,0],supercell[0,1],supercell[0,2],
                supercell[1,0],supercell[1,1],supercell[1,2],
                supercell[2,0],supercell[2,1],supercell[2,2]))
    phonon = ph.Phonopy(unitcell, supercell, factor=ph.units.VaspToTHz,symprec=symprec)
    phonon.generate_displacements(distance=disp)
    
    scstruct = phonon.get_supercell()
    supercells = phonon.get_supercells_with_displacements()
    
    numcells = int(round(np.linalg.det(supercell)))
    
    # here it is required to create a new template for the supercell
    # for this the standard ordering for supercells in phonopy is used
    if atom_template is not None:
        new_atom_types = []
        for aid, atype in enumerate(atom_types):
            for sid in range(numcells):
                new_atom_types.append(atype)
        atom_types = new_atom_types
    # here, do the same thing for atomic charges
    if charges is not None:
        chdata = []
        for aid, coor in enumerate(testgeom.coors):
            for sid in range(numcells):
                chdata.append(testgeom.charges[aid])
        chdata = np.array(chdata)
        
    
    
    # initialize lammps for the supercell
    scgeom = geometry.geo(".", filetype="direct", elems=np.array(scstruct.get_chemical_symbols()), 
                          coors=scstruct.get_positions(), cell=scstruct.get_cell(),
                          atom_types=atom_types)
    
    unrotated_cell = copy.copy(scgeom.cell)
    
    
    scgeom.convert_cell_to_lammps()
    mli = mlip_lammps_interface(escut=escut)
    if charges is None:
        mli.load_mlip_potential(potfile, scgeom, uelemsort=uelemsort)
    else:
        mli.load_hybrid_mlip_charge_potential(potfile, scgeom, charge_data=chdata, charge_style=charge_type)
    
    if (not silent) & (rank == 0):
        print("computing displacements")
    force_sets = []
    for sc in tqdm(supercells):
        rotated_coors = util.rotate_vec(sc.get_positions(), unrotated_cell, scgeom.cell)
        mli.update_coors(scgeom.cell,rotated_coors)
        forces = mli.calc_force()
        forces = util.rotate_vec(forces, scgeom.cell, unrotated_cell)
        force_sets.append(forces)
        
    phonon.set_forces(np.array(force_sets))
    
    if (not silent) & (rank == 0):
        print ("computing force constants... ")
    phonon.produce_force_constants()
    gammafreqs = phonon.get_frequencies([0, 0, 0])
    if rank == 0:
        with open("gammafreqs.dat", "w") as freqfile:
            for val in gammafreqs:
                freqfile.write("%14.8f  %14.8f\n" % (val, val/0.0299792458))
                if val < -0.1:
                    if not silent:
                        print ("WARNING: imaginary frequency " + str(val))
            freqfile.close()
            
        fcs = phonon.get_force_constants()
    
        if txtFC:
            ph.file_IO.write_FORCE_CONSTANTS(fcs)
        else:
            ph.file_IO.write_force_constants_to_hdf5(fcs)
        
    
    
    return


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="computes the phonon band structure with mlip mtp potentials")


    parser.add_argument("-filetype", dest="filetype", type=str, default="POSCAR", help="filetype of the structures")
    parser.add_argument("-potfile", dest="potfile", type=str, default="pot.mtp", help="name of the file that stores the FFP")
    parser.add_argument("-disp", dest="disp", type=float, default=0.01, help="displacement distance")
    parser.add_argument("-symprec", dest="symprec", type=float, default=1e-5, help="symmetry precision for phonopy")
    parser.add_argument("-dim", dest="dim", type=int, nargs="*", default=[1,1,1], help="supercell size for the phonon calculation; 3 or 9 arguments")
    parser.add_argument("-template", dest="atom_template", type=str, default=None, 
                          help="provide a template for the atom types used in a MLP type MTP in the form of a cfg file. The atoms have to be sorted in the same way as in the reference geometries. The masses will be taken from the actual elements")
    parser.add_argument("-charges", dest="charges", type=str, default=None, 
                          help="provide atomic charges to be used in addition to the MTP. The charges need to be ordered in the same way as in the structure file.")
    parser.add_argument("-charge_type", dest="charge_type", type=str, default="gauss/dsf", help="defines which type of charges are applied. The default are Gaussian charges")
    parser.add_argument("-escut", dest="escut", type=float, default=None, 
                        help="low-range electrostatic cutoff")
    parser.add_argument("-relax", dest="relax", action="store_true", default=False, help="perform relaxation of the box first")
    parser.add_argument("-silent", dest="silent", action="store_true", default=False, help="no text output")
    parser.add_argument("-relaxint", dest="relaxint", action="store_true", default=False, help="perform relaxation of the internal coordinates first")
    parser.add_argument("-txtFC", dest="txtFC", action="store_true", default=False, help="write force constants as text file instead of hdf5 (default)")
    parser.add_argument("-uelemsort", dest="uelemsort", action="store_true", default=False, help="when trying to assign atom types they will be sorted according to the alphabetical ordering of the element name instead of according to which atom types comes first in the atom list. If you get really crazy results this might be worth trying")

    parser.add_argument("filename", type=str, help="path to the relaxed structures")


    args = parser.parse_args()
    
    if rank==0:
        afp = open("p_cmnd.log","a")
        afp.write(" ".join(sys.argv)+"\n")
        afp.close()



    compute_mlip_band(**vars(args))
    
    commw.Disconnect 