#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
This script serves to compute the elastic moduli using lammps with the MLIP interface

"""



import argparse
import sys
import numpy as np
import geometry
import copy
from ML_FF_aux.mlip_lammps_interface import mlip_lammps_interface
import swfileIO.ml_abn as ml_abn
from tqdm import tqdm
import util

from mpi4py import MPI

commw = MPI.COMM_WORLD
rank = commw.Get_rank()
nprocs = commw.Get_size()

def perform_deformation(cell, direction, indices, dist):
    """
      function to perform a strain operation on a cell
    """
    # have to change off-diagonals for xx and yy
    
    strain_matrix = np.zeros((3,3))
    strain_matrix[indices[0],indices[1]] = dist
    strain_matrix[indices[1],indices[0]] = dist
#    print(strain_matrix)
    delta = dist
    if indices[0] != indices[1]:
        # I am wondering - is this wrong? Does this maybe only work for symmetric cell matrices?
        # but for MOF-74 it did seem to work and there the cell matrix is not symmetric
        delta *= 2 # apply twice the strain for off-diagonal elements
#    print(delta)
#    print(np.dot(cell,strain_matrix))
#    print(cell)
    cell = np.dot(cell, np.eye(3)+strain_matrix)
#    print(cell)
    return cell, delta

def compute_mlip_elastic(filename, filetype="POSCAR", potfile="pot.mtp", atom_template=None, 
                         dist=1e-6, relax=False, maxiter=1000, like_vasp=False, charges=None, 
                         charge_type="gauss/dsf", absolute=False, uelemsort=False):
    
    
    # for metal units - to obtain values in GPa
    c_factor = 1e-4
    
    # load lammps interface
    if atom_template is not None:
        mobj = ml_abn.ml_abn(".",filetype="none")
        mobj.read_MLIP_cfg(atom_template)
        atom_types = np.array(mobj.elems[0])
    else:
        atom_types = None
    testgeom = geometry.geo(filename, filetype=filetype, atom_types=atom_types)
    init_geom = copy.deepcopy(testgeom)
    testgeom.convert_cell_to_lammps()
    rotated_cell = copy.copy(testgeom.cell)
    mli = mlip_lammps_interface()
    if charges is None:
        mli.load_mlip_potential(potfile, testgeom, uelemsort=uelemsort)
    else:
        mli.load_hybrid_mlip_charge_potential(potfile, testgeom, charges, charge_style=charge_type)
    
    # relax the box
    if relax:
        mli.relax_box()
        testgeom.coors = mli.get_xyz()
        testgeom.cell = mli.get_cell()
        init_geom.coors = util.rotate_vec(testgeom.coors, rotated_cell, init_geom.cell)
        init_geom.cell  = util.rotate_vec(testgeom.cell, rotated_cell, init_geom.cell)
    
    mli.calc_energy()
    initial_stress = mli.get_stress_tensor()
    initial_stress = util.rotate_tens(initial_stress, testgeom.cell, init_geom.cell)
    
    deriv_tot = []
    straindiffs = []
    # loop over the 6 non-zero cell parameters and apply a positive and negative distortion
    indices = [[0,0],[1,1],[2,2],[1,0],[2,0],[2,1]]
    for direction in tqdm(range(6)): 
#        l0 = init_geom.cell[indices[direction][0],indices[direction][0]]
#        if np.abs(l0) > 0:
        # apply strain on the ORIGINAL cell
        if not absolute:
            newcell, delta = perform_deformation(copy.copy(init_geom.cell), direction, indices[direction], dist)
        else:
            newcell, delta = perform_absolute_deformation(copy.copy(init_geom.cell), direction, indices[direction], dist)
        testgeom.cart_to_frac()
        testgeom.cell = newcell
        testgeom.frac_to_cart()
        before_rot = copy.copy(testgeom.cell)
        testgeom.convert_cell_to_lammps()
        mli.update_cell(testgeom.cell)
        if not like_vasp:
            mli.relax_int(maxiter=maxiter)
        else:
            mli.calc_energy()  # do a single point calculation
        stress = mli.get_stress_tensor()
        # rotate back into initial coordinate system
        stress = util.rotate_tens(stress,testgeom.cell,before_rot)
        deriv1 = -(stress-initial_stress)/delta*c_factor
        
        # apply strain on the ORIGINAL cell
        newcell, delta = perform_deformation(copy.copy(init_geom.cell), direction, indices[direction], -dist)
        testgeom.cart_to_frac()
        testgeom.cell = newcell
        testgeom.frac_to_cart()
        before_rot = copy.copy(testgeom.cell)
        testgeom.convert_cell_to_lammps()
        mli.update_cell(testgeom.cell)
        if not like_vasp:
            mli.relax_int(maxiter=maxiter)
        else:
            mli.calc_energy()  # do a single point calculation
        stress = mli.get_stress_tensor()
        # rotate back into initial coordinate system
        stress = util.rotate_tens(stress,testgeom.cell,before_rot)
        deriv2 = -(stress-initial_stress)/delta*c_factor
#        else:
#            # to avoid a division by 0 error
#            deriv1 = np.zeros((3,3))
#            deriv2 = np.zeros((3,3))
        straindiffs.append(np.sum(np.abs(deriv2-deriv1)))
        
#        print((deriv1+deriv2)/2)
        # rotate back the derivatives to be consistent with the original cell definition
#        rot_deriv = util.rotate_tens((deriv1+deriv2)/2,testgeom.cell,initial_cell)
#        print(rot_deriv)
        rot_deriv = (deriv1+deriv2)/2
        deriv_tot.append(rot_deriv)
    
    deriv_tot=np.array(deriv_tot)
    
    if rank == 0:
        print("absolute sum of differences of the derivative between positive and negative strain:")
        for sdiff in straindiffs:
            print(sdiff)
        print("these numbers should be as close to zero as possible. Modify the dist parameter to converge!")
    
    # get the actual elastic tensor elements
    # WARNING: index nightmare
    elastic = np.zeros((6,6))
    
    elastic[0,0] = deriv_tot[0][0,0]
    elastic[1,1] = deriv_tot[1][1,1]
    elastic[2,2] = deriv_tot[2][2,2]
    
    elastic[0,1] = elastic[1,0] = (deriv_tot[0][1,1]+deriv_tot[1][0,0])/2
    elastic[0,2] = elastic[2,0] = (deriv_tot[0][2,2]+deriv_tot[2][0,0])/2
    elastic[1,2] = elastic[2,1] = (deriv_tot[1][2,2]+deriv_tot[2][1,1])/2
    
    elastic[3,3] = deriv_tot[3][1,0]
    elastic[4,4] = deriv_tot[4][2,0]
    elastic[5,5] = deriv_tot[5][2,1]
    
    elastic[0,3] = elastic[3,0] = (deriv_tot[0][1,0]+deriv_tot[3][0,0])/2
    elastic[0,4] = elastic[4,0] = (deriv_tot[0][2,0]+deriv_tot[4][0,0])/2
    elastic[0,5] = elastic[5,0] = (deriv_tot[0][2,1]+deriv_tot[5][0,0])/2
    
    elastic[1,3] = elastic[3,1] = (deriv_tot[1][1,0]+deriv_tot[3][1,1])/2
    elastic[1,4] = elastic[4,1] = (deriv_tot[1][2,0]+deriv_tot[4][1,1])/2
    elastic[1,5] = elastic[5,1] = (deriv_tot[1][2,1]+deriv_tot[5][1,1])/2
    
    elastic[2,3] = elastic[3,2] = (deriv_tot[2][1,0]+deriv_tot[3][2,2])/2
    elastic[2,4] = elastic[4,2] = (deriv_tot[2][2,0]+deriv_tot[4][2,2])/2
    elastic[2,5] = elastic[5,2] = (deriv_tot[2][2,1]+deriv_tot[5][2,2])/2
    
    elastic[3,4] = elastic[4,3] = (deriv_tot[3][2,0]+deriv_tot[4][1,0])/2
    elastic[3,5] = elastic[5,3] = (deriv_tot[3][2,1]+deriv_tot[5][1,0])/2
    elastic[4,5] = elastic[5,4] = (deriv_tot[4][2,1]+deriv_tot[5][2,0])/2

#    print("total derivatives:")
#    print(deriv_tot)
    
    if rank == 0:
        # here write a generic output file
        with open("elastic_m.dat","w") as ofp:
            print("elastic moduli:")
            ofp.write("# %12s %12s %12s %12s %12s %12s %s\n" % ("XX","YY","ZZ","XY","XZ","YZ","GPa"))
            for line in elastic:
                print("  %12.8f %12.8f %12.8f %12.8f %12.8f %12.8f" % (line[0],line[1],line[2],
                                                                     line[3],line[4],line[5]))
                ofp.write("  %12.8f %12.8f %12.8f %12.8f %12.8f %12.8f\n" % (line[0],line[1],line[2],
                                                                         line[3],line[4],line[5]))
            print("values given in GPa")
            ofp.close()
        
        # here, write an OUTCAR in the format of VASP to be processed with Tomas' Gruneisen scripts
        # this has to be in kbar
        with open("OUTCAR_elastic","w") as ofp:
            GPa2kbar = 10
            ofp.write(" TOTAL ELASTIC MODULI (kBar)\n")
            ofp.write(" %10s %12s %12s %12s %12s %12s %12s \n" % ("Direction","XX","YY","ZZ","XY","YZ","ZX"))
            ofp.write(" ------------------------------------------------------------------------------------------\n")
            direcs = ("XX","YY","ZZ","XY","YZ","ZX")
            for eid in range(6):
                # need to swap the latter two entries
                if eid == 4:
                    ind = 5
                elif eid == 5:
                    ind = 4
                else:
                    ind = eid
                line = elastic[ind]*GPa2kbar
                ofp.write(" %10s %12.8f %12.8f %12.8f %12.8f %12.8f %12.8f\n" % (direcs[eid],
                                                                                 line[0],line[1],line[2],
                                                                                 line[3],line[5],line[4]))
            ofp.write(" ------------------------------------------------------------------------------------------\n")
            ofp.close()
            
    
    
        
#        cparam = testgeom.cell[indices[direction][0],indices[direction][1]]
#        deform = 
    return elastic


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="computes elastic constants with mlip mtp potentials")


    parser.add_argument("-filetype", dest="filetype", type=str, default="POSCAR", help="filetype of the structures")
    parser.add_argument("-potfile", dest="potfile", type=str, default="pot.mtp", help="name of the file that stores the FFP")
    parser.add_argument("-dist", dest="dist", type=float, default=1e-4, help="deformation size for the applied strains given as a fraction (default: 1e-4)")
    parser.add_argument("-absolute", dest="absolute", action="store_true", default=False, help="Use an absolute strain given in A in the dist parameter instead of the fraction of a strain. NOT IMPLEMENTED YET!")
    parser.add_argument("-relax", dest="relax", action="store_true", default=False, help="perform relaxation of the box first")
    parser.add_argument("-charges", dest="charges", type=str, default=None, 
                          help="provide atomic charges to be used in addition to the MTP. The charges need to be ordered in the same way as in the structure file.")
    parser.add_argument("-charge_type", dest="charge_type", type=str, default="gauss/dsf", help="defines which type of charges are applied. The default are Gaussian charges")
    parser.add_argument("-like_vasp", dest="like_vasp", action="store_true", default=False, help="carry out the calculation as VASP does - so with no relaxation of the internal coordinates for the strained cell. NOT IMPLEMENTED PROPERLY YET!")
    parser.add_argument("-maxiter", dest="maxiter", type=int, default=1000, help="maximum number of iterations for relaxation")
    parser.add_argument("-template", dest="atom_template", type=str, default=None, 
                          help="provide a template for the atom types used in a MLP type MTP in the form of a cfg file. The atoms have to be sorted in the same way as in the reference geometries. The masses will be taken from the actual elements")
    parser.add_argument("-uelemsort", dest="uelemsort", action="store_true", default=False, help="when trying to assign atom types they will be sorted according to the alphabetical ordering of the element name instead of according to which atom types comes first in the atom list. If you get really crazy results this might be worth trying")

    parser.add_argument("filename", type=str, help="path to the relaxed structures")


    args = parser.parse_args()
    
    if rank==0:
        afp = open("p_cmnd.log","a")
        afp.write(" ".join(sys.argv)+"\n")
        afp.close()



    compute_mlip_elastic(**vars(args))
    
    commw.Disconnect    


