#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
 This script serves to automatically compute Gruneisen Parameters from a MLIP style MTP
"""

import argparse
import sys
from ML_FF_aux.compute_mlip_elastic import compute_mlip_elastic, perform_deformation
from ML_FF_aux.compute_mlip_band import compute_mlip_band
import os
import util
import numpy as np
import copy
import geometry
from QHA_Tomas.gruneisen_tensor import gruneisen_tensor
import time

from mpi4py import MPI

commw = MPI.COMM_WORLD
rank = commw.Get_rank()
nprocs = commw.Get_size()

def apply_strain(cell,DIM):
    # function to apply a strain on a cell
    if len(args.eps)==3:
        appleps=np.diag([float(DIM[0]),float(DIM[1]),float(DIM[2])])
    elif len(args.eps)==9:
        appleps=np.array([[float(DIM[0]),float(DIM[1]),float(DIM[2])],
                          [float(DIM[3]),float(DIM[4]),float(DIM[5])],
                          [float(DIM[6]),float(DIM[7]),float(DIM[8])]])
    cellstrained=np.dot(cell,np.eye(3)+appleps)
    
    return cellstrained

def rprint(text):
    if rank == 0:
        print(text)
            

def compute_mlip_gruneisen(filename, filetype="POSCAR", potfile="pot.mtp", atom_template=None, 
                         eldist=1e-6, disp=0.001, stdist=[1e-3], dim=[1,1,1], mesh=[10,10,10],
                         iso=False, xxz=False):
    
    # steps in the procedure: 
    # - compute elastic constants
    # - compute phonon bands for unrestrained structure
    # - strain cell in all directions
    # - compute phonon bands for strained cells
    # - compute Gruneisen parameters with Tomas' scripts
    
    if nprocs > 1:
        print("WARNING: parallelization is still suffering from problems currently")
        print("for reliable results run in serial!")
    
    
    #### ELASTIC  ####
    if not os.path.exists("OUTCAR_elastic"):
        rprint("calculating elastic constants...")
        compute_mlip_elastic(filename, filetype=filetype, potfile=potfile, atom_template=atom_template,
                             dist=eldist)
    else:
        rprint("found elastic constants!")
    if rank == 0:
        util.mkdir("Elastic")
    while not os.path.exists("Elastic"):
        time.sleep(0.001)
    if rank == 0:
        os.system("cp OUTCAR_elastic Elastic/OUTCAR")
    
    
    #### UNSTRAINED BAND STRUCTURE ####
    if not os.path.exists("equi"):
        rprint("calculating undisplaced band structure...")
        compute_mlip_band(filename, filetype=filetype, potfile=potfile, atom_template=atom_template,
                          disp=disp, dim=dim)
        if rank == 0:
            util.mkdir("equi")
        while not os.path.exists("equi"):
            time.sleep(0.001)
        if rank == 0:
            os.system("cp "+filename+" force_constants.hdf5 equi")
    else:
        rprint("found undisplaced band structure!")
        
    
    
    #### STRAIN BAND STRUCTURE ####
    # for each of the distances - also need to do +/-
    all_dists = []
    for strain_dist in stdist:
        all_dists.append(strain_dist)
        all_dists.append(-strain_dist)
    
    geom = geometry.geo(filename,filetype=filetype)
    if iso:
        directions = [[0,0]]
        dir_names = ["iso"]
    elif xxz:
        directions = [[0,0],[2,2]]
        dir_names = ["XX","ZZ"]
    else:
        directions = [[0,0],[1,1],[2,2],[1,0],[2,0],[2,1]]
        dir_names = ["XX","YY","ZZ","XY","XZ","YZ"]
    disp_paths = {}
    for dname in dir_names:
        disp_paths[dname] = []
    for dist in all_dists:
        for did,direc in enumerate(directions):
            dirstr = dir_names[did] + "_" + str(dist)
            if not os.path.exists(dirstr):
                rprint("calculating strained cell "+dirstr+"...")
                
                if iso:
                    newcell = copy.copy(geom.cell)*(1+dist)
                elif xxz:
                    
                    if dir_names[did] == "ZZ":
                        newcell, delta = perform_deformation(copy.copy(geom.cell),did,direc,dist)
                    else:
                        newcell = copy.copy(geom.cell)
                        newcell[:,0] *= (1+dist)
                        newcell[:,1] *= (1+dist)
                else:
                    newcell, delta = perform_deformation(copy.copy(geom.cell),did,direc,dist)
                if rank == 0:
                    util.mkdir(dirstr)
                while not os.path.exists(dirstr):
                    time.sleep(0.001)
                thisgeom = copy.deepcopy(geom)
                thisgeom.cart_to_frac()
                thisgeom.set_cell(newcell)
                thisgeom.frac_to_cart()
                
                thisgeom.write_file(dirstr+"/POSCAR", filetype="POSCAR")
                compute_mlip_band(dirstr+"/POSCAR", filetype="POSCAR", potfile=potfile, atom_template=atom_template,
                                  disp=disp, dim=dim, relaxint=True)
                if rank == 0:
                    os.system("cp force_constants.hdf5 "+dirstr)
                    os.system("cp POSCAR_relaxed "+dirstr+"/POSCAR")
            else:
                rprint("found band structure calculation for strained cell "+dirstr+"!")
            disp_paths[dir_names[did]].append(dirstr)
                
                
    
    
    
    #### use Tomas' script to compute Gruneisen parameters ####
    if rank == 0: # I doubt this is properly parallelized
        if iso:
            gruneisen_tensor(aDIM=dim, amode=["mesh"], amesh=mesh, 
                             apath11=disp_paths["iso"], apath22=disp_paths["iso"], apath33=disp_paths["iso"], 
                             acompliancepath="Elastic/", apath0="equi", asavepath=".", noshow=True,
                             afactor11="1/3", afactor22="1/3", afactor33="1/3") 
        elif xxz:
            gruneisen_tensor(aDIM=dim, amode=["mesh"], amesh=mesh, 
                             apath11=disp_paths["XX"], apath22=disp_paths["XX"], apath33=disp_paths["ZZ"], 
                             acompliancepath="Elastic/", apath0="equi", asavepath=".", noshow=True,
                             afactor11="1/2", afactor22="1/2") 
        else:
            gruneisen_tensor(aDIM=dim, amode=["mesh"], amesh=mesh, 
                             apath11=disp_paths["XX"], apath22=disp_paths["YY"], apath33=disp_paths["ZZ"], 
                             apath12=disp_paths["XY"], apath13=disp_paths["XZ"], apath23=disp_paths["YZ"],
                             acompliancepath="Elastic/", apath0="equi", asavepath=".", noshow=True) 
    
    
    
    return

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="computes Gruneisen parameters with mlip mtp potentials using the scripts by Tomas Kamencek")


    parser.add_argument("-filetype", dest="filetype", type=str, default="POSCAR", help="filetype of the structures")
    parser.add_argument("-potfile", dest="potfile", type=str, default="pot.mtp", help="name of the file that stores the FFP")
    parser.add_argument("-eldist", dest="eldist", type=float, default=1e-3, help="deformation size for elastic constants")
    parser.add_argument("-disp", dest="disp", type=float, default=0.01, help="displacement distance")
    parser.add_argument("-stdist", dest="stdist", type=float, default=[1e-3], nargs="*", help="deformation size for Gruneisen evaluation. Accepts 1 or 2 parameters - if 2 parameters are given a 4th order finite difference scheme is applied")
    parser.add_argument("-iso", dest="iso", action="store_true", default=False, help="Compute for isotropic system (some factors are still wrong)")
    parser.add_argument("-xxz", dest="xxz", action="store_true", default=False, help="Compute for equal strains in x and y and a different strain in z. No non-diagonal strains (like for MOF-74)")
    parser.add_argument("-dim", dest="dim", type=int, nargs="*", default=[1,1,1], help="supercell size for the phonon calculation; 3 or 9 arguments")
    parser.add_argument("-mesh", dest="mesh", type=int, nargs=3, default=[10,10,10], help="mesh density for evaluation")
    parser.add_argument("-template", dest="atom_template", type=str, default=None, 
                          help="provide a template for the atom types used in a MLP type MTP in the form of a cfg file. The atoms have to be sorted in the same way as in the reference geometries. The masses will be taken from the actual elements")

    parser.add_argument("filename", type=str, help="path to the relaxed structures")


    args = parser.parse_args()
    
    if rank==0:
        afp = open("p_cmnd.log","a")
        afp.write(" ".join(sys.argv)+"\n")
        afp.close()



    compute_mlip_gruneisen(**vars(args))
    
    commw.Disconnect 