#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import argparse
import sys
import swfileIO.dump
import numpy as np
import util
import geometry

def convert_matrix_to_cif(mat,geom):
    A = geom.cell
    rcell = geom.get_reciprocal_lattice_vectors()
    N = np.zeros((3,3))
    for i in range(3):
        N[i,i] = np.linalg.norm(rcell[i])/2/np.pi
    Q = np.matmul(A,N)
    invQ = np.linalg.inv(Q)
    #print(Q)
    return np.matmul(np.matmul(invQ,mat),Q)

def write_displacement_ellipsoid_cif(fname,geom,Umat):
    # very primitive cif file writer - use at your own risk! VESTA can visualize the file
    # the Umat matrix first needs to be transformed from Cartesian coordinates 
    #print(Umat[0])
    Umat_cif = convert_matrix_to_cif(Umat,geom)
    #print(Umat_cif[0])
    rad2deg = 360/2/np.pi
    geom.cart_to_frac()
    ofp = open(fname,"w")
    ofp.write("# Computed from MD trajectory\n")
    ofp.write("\n")
    ofp.write("_chemical_name_common       'VASP_MD_traj'\n")
    lengths, angles = geom.get_cell_lengths_and_angles()
    ofp.write("_cell_length_a              %12.8f\n" % (lengths[0]))
    ofp.write("_cell_length_b              %12.8f\n" % (lengths[1]))
    ofp.write("_cell_length_c              %12.8f\n" % (lengths[2]))
    ofp.write("_cell_angle_alpha           %12.8f\n" % (angles[0]*rad2deg))
    ofp.write("_cell_angle_beta            %12.8f\n" % (angles[1]*rad2deg))
    ofp.write("_cell_angle_gamma           %12.8f\n" % (angles[2]*rad2deg))
    ofp.write("_cell_volume                %12.8f\n" % (geom.get_volume()))
    ofp.write("_space_group_name_H-M_alt   'P 1'\n")
    ofp.write("_space_group_IT_number      1\n")
    ofp.write("\n")
    ofp.write("loop_\n")
    ofp.write("_space_group_symop_operation_xyz\n")
    ofp.write("   'x, y, z'\n")
    ofp.write("\n")
    ofp.write("loop_\n")
    ofp.write("   _atom_site_label\n")
    ofp.write("   _atom_site_type_symbol\n")
    ofp.write("   _atom_site_fract_x\n")
    ofp.write("   _atom_site_fract_y\n")
    ofp.write("   _atom_site_fract_z\n")
    uindicators = []
    elemnumdict = {}
    for aid,coor in enumerate(geom.coors):
        if geom.elems[aid] not in elemnumdict.keys():
            elemnumdict[geom.elems[aid]] = 1
        else:
            elemnumdict[geom.elems[aid]] += 1
        uindicators.append(geom.elems[aid]+str(elemnumdict[geom.elems[aid]]))
        ofp.write("   %6s %3s %12.8f %12.8f %12.8f\n" % (uindicators[-1],geom.elems[aid],coor[0],coor[1],coor[2]))
    ofp.write("loop_\n")
    ofp.write("   _atom_site_aniso_label\n")
    ofp.write("   _atom_site_aniso_U_11\n")
    ofp.write("   _atom_site_aniso_U_22\n")
    ofp.write("   _atom_site_aniso_U_33\n")
    ofp.write("   _atom_site_aniso_U_12\n")
    ofp.write("   _atom_site_aniso_U_13\n")
    ofp.write("   _atom_site_aniso_U_23\n")
    for aid,mat in enumerate(Umat_cif):
        ofp.write("   %6s %12.8f %12.8f %12.8f %12.8f %12.8f %12.8f\n" % (uindicators[aid],mat[0,0],mat[1,1],mat[2,2],mat[0,1],mat[0,2],mat[1,2]))
    ofp.write("\n")
    ofp.write("\n")


def eval_thermal_displacements(fname):
    dumpdata = swfileIO.dump.dump(fname,filetype="vasp")
    dumpdata.center_coordinates()
    coors= dumpdata.coors
    
    # the problem with averaging like this are the periodic boundary conditions - they would have to be considered
    # or a feature for unwrapping the trajectories and consider all numbers without boundary jumps
    avg_positions = np.mean(coors,axis=0)
    avg_cell = np.mean(dumpdata.cells,axis=0)
    util.write_POSCAR("POSCAR_mean",avg_cell,dumpdata.elems,avg_positions)
    geom = geometry.geo("POSCAR")
    
    Umat = np.zeros((len(avg_positions),3,3))
    for tid, coor in enumerate(coors):
        for i in range(3):
            for j in range(3):
                Umat[:,i,j] += (coors[tid,:,i]-avg_positions[:,i])*(coors[tid,:,j]-avg_positions[:,j])
    Umat /= len(coors)
    
    ofp = open("Umat.dat","w")
    for eid,entry in enumerate(Umat):
        ofp.write("# Atom %d %s\n" % (eid,dumpdata.elems[eid]))
        for i in range(3):
            ofp.write("%12.8f %12.8f %12.8f\n" % (entry[i,0],entry[i,1],entry[i,2]))
    ofp.close()
    
    geom.coors = avg_positions
    geom.cell = avg_cell
    write_displacement_ellipsoid_cif("MD_Umat.cif",geom,Umat)
    #print(avg_positions - geom.coors)
    
    return


if __name__ == "__main__":
    
    
    parser = argparse.ArgumentParser(description="This script serves to compute the thermal displacements of atoms in an MD run")

    parser.add_argument("-fname", dest="fname", type=str, default=".", help="directory of the files")


    args = parser.parse_args()
    
    afp = open("p_cmnd.log","a")
    afp.write(" ".join(sys.argv)+"\n")
    afp.close()
    
    eval_thermal_displacements(**vars(args))
