#!/usr/bin/python3
# -*- coding: utf-8 -*-

# this class provides an interface of lammps to the mlip potential
# credit for certain methods goes to Johannes P. Dürholt and the pylmps interface

from lammps import lammps
import numpy as np
from mpi4py import MPI
import aprops.elems as elems

commw = MPI.COMM_WORLD
rank = commw.Get_rank()
nprocs = commw.Get_size()



class mlip_lammps_interface():
    
    def __init__(self, escut=None, extra_bonds=20):
        """
          escut ... cutoff for optional electrostatic interactions
          extra_bonds ... the number of extra bonds that will be reserved for the dynamically assigned bonds
        """
        self.lmps = None
        self.escut = escut
        self.extra_bonds = extra_bonds
        return
    
    def load_mlip_potential(self,potfile,init_geom,uelemsort=False):
        """
        loads mlip and prepares for performing single point calculations
        uses a geometry object as an input and requires the name of the mlip potential file
        uelemsort ... when trying to assign atom types they will be sorted according to the alphabetical ordering of the element name instead of according to which atom types comes first in the atom list.
        """
        if self.lmps is None:
            lmps = lammps(cmdargs=["-screen","none"], comm=commw)
            self.lmps = lmps
        else:
            lmps = self.lmps
        lmps.command("units metal")
        if self.escut is None:
            lmps.command("atom_style atomic")
            bnd_num = None
        else:
            lmps.command("atom_style bond")
            bnd_num = 1
        lmps.command("boundary p p p")
        lmps.command("box tilt large")
        lmps.command("atom_modify map yes")  # this is necessary for atomic styles otherwise scatter atoms does not work
        # HACK here, we write to a file from which we read to initialize lammps - which is not exactly pretty
        init_geom.write_file("tmp.pos",filetype="lmp_atomic",bond_num=bnd_num,uelemsort=uelemsort)
        self._read_lmp_special()
        #lmps.set_cell()
        #lmps.scatter_atoms("x",1,3,np.ctypeslib.as_ctypes(init_geom.coors))
        with open("mlip.ini","w") as ofp:
            ofp.write("mtp-filename    "+potfile)
        
        lmps.command("pair_style mlip mlip.ini")
        lmps.command("pair_coeff * *")
        
        self.geom = init_geom
        self._post_setup(lmps)
        
#        lmps.command("run 0")
        self.natoms = len(self.geom.coors)
        return self.lmps
    
    def load_hybrid_mlip_charge_potential(self,potfile,init_geom,chargefile=None,charge_style="gauss/dsf",
                                                   kspace=False,kspace_prec=5e-6,cutoff=12,alpha=0.05,
                                                   charge_data=[]):
        """
         method to create an FF with electrostatic interactions on top of a mlip potential 
         potfile ... file name for the mlip potential
         init_geom ... geometry object for the initial geometry
         chargefile ... file name for the atomic charges (as a simple list)
         charge_style ... which charge style should be used
         kspace ... evaluate long range electrostatic interactions (with the much faster pppm solver) 
         kspace_prec ... precision for the kspace solver
         cutoff ... cutoff of electrostatic interactions in Angstroms
         alpha ... alpha parameter used by some charge styles
         charge_data ... use atomic charges from an array instead
        """
        if self.lmps is None:
            lmps = lammps(cmdargs=["-screen","none"], comm=commw)
            self.lmps = lmps
        else:
            lmps = self.lmps
        lmps.command("units metal")
        if self.escut is None:
            lmps.command("atom_style charge")
            bnd_num = None
        else:
            lmps.command("atom_style hybrid charge bond")
            bnd_num = 1
        lmps.command("boundary p p p")
        lmps.command("box tilt large")
        lmps.command("atom_modify map yes")  # this is necessary for atomic styles otherwise scatter atoms does not work
        if chargefile is not None:
            charges = np.genfromtxt(chargefile).ravel()
            assert len(init_geom.coors) == len(charges), "ERROR: number of charges provided is not equal the number of atoms in the system"
            init_geom.set_charges(charges)
        else:
            assert len(init_geom.coors) == len(charge_data), "ERROR: number of charges provided is not equal the number of atoms in the system"
            init_geom.set_charges(charge_data)
        # HACK here, we write to a file from which we read to initialize lammps - which is not exactly pretty
        init_geom.write_file("tmp.pos",filetype="lmp_charge",bond_num=bnd_num)
        self._read_lmp_special()
        #lmps.set_cell()
        #lmps.scatter_atoms("x",1,3,np.ctypeslib.as_ctypes(init_geom.coors))
        with open("mlip.ini","w") as ofp:
            ofp.write("mtp-filename    "+potfile)
        
        if charge_style == "gauss/dsf":
            lmps.command("pair_style hybrid/overlay mlip mlip.ini buck6d/coul/gauss/dsf     0.9000     0.9000    %12.6f" % (cutoff))
        elif charge_style == "coul/cut":
            lmps.command("pair_style hybrid/overlay mlip mlip.ini coul/cut  %12.6f" % (cutoff))
        elif charge_style == "coul/dsf":   # dsf has a smooth decay to zero - so it should be superior - it also has the erf definition however maybe there is not such a large difference compared to the buck6d potential - but the alpha parameter is now global - if it is very low it should just behave like regular point charges
            lmps.command("pair_style hybrid/overlay mlip mlip.ini coul/dsf  %12.6f  %12.6f" % (alpha, cutoff))
        elif charge_style == "coul/wolf":
            lmps.command("pair_style hybrid/overlay mlip mlip.ini coul/wolf  %12.6f  %12.6f" % (alpha, cutoff))
        else:
            assert False, "ERROR: charge style %s is not known" % charge_style
            
        lmps.command("pair_coeff * * mlip")
        
        
        if charge_style == "gauss/dsf":
            # here need to add the pair coefficients for the rest of the potential
            # the question here is how to obtain the parameters - is it reasonable to have them in a python data file as a dictionary?
            # I cannot only use the elems here - as different atom type based potentials might be used
            if init_geom.atom_types is not None:
                syselems = init_geom.atom_types
            else:
                syselems = init_geom.elems
            uelems,uinds = np.unique(syselems,return_index=True)
            logical = np.argsort(uinds)
            uelems = uelems[logical]
            uinds = uinds[logical]  # to get the actual element
            for eid1, at1 in enumerate(uelems):
                for eid2, at2 in enumerate(uelems):
                    if eid2 >= eid1:  # interactions are symmetric
                        elem1 = init_geom.elems[uinds[eid1]]
                        elem2 = init_geom.elems[uinds[eid2]]
                        sigma1 = elems.sigmas[elem1.lower()]
                        sigma2 = elems.sigmas[elem2.lower()]
                        # mixing is not concerning in this case because this is mostly relevant for the close
                        # range interactions
                        sigma_pair = (sigma1**2 + sigma2**2)**0.5 # mixing rule
                        # no vdW - only coulomb
                        lmps.command("pair_coeff     %d     %d   buck6d/coul/gauss/dsf   0   0   0   0      %12.8f" %
                                     (eid1+1,eid2+1,sigma_pair))
        else:
            lmps.command("pair_coeff * * %s" % charge_style) # for all styles that do not have atom type dependent definitions
            
            
        self.geom = init_geom
        self._post_setup(lmps)
        self.natoms = len(self.geom.coors)
        return self.lmps 
    
    def load_gaussian_charge_potential(self,init_geom,chargefile,kspace=False,kspace_prec=5e-6,cutoff=12):
        """
         method to create an FF with electrostatic gaussian interactions only
         init_geom ... geometry object for the initial geometry
         chargefile ... file name for the atomic charges (as a simple list)
         kspace ... evaluate long range electrostatic interactions (with the much faster pppm solver) 
         kspace_prec ... precision for the kspace solver
         cutoff ... cutoff of electrostatic interactions in Angstroms
        """
        return self.load_charge_potential(init_geom,chargefile,kspace=kspace,kspace_prec=kspace_prec,
                                      cutoff=cutoff,charge_style="gauss/dsf")
         
         
    def load_charge_potential(self,init_geom,chargefile,kspace=False,kspace_prec=5e-6,cutoff=12,
                              charge_style="gauss/dsf",alpha=0.05):
        """
         method to create an FF with electrostatic interactions only
         init_geom ... geometry object for the initial geometry
         chargefile ... file name for the atomic charges (as a simple list)
         kspace ... evaluate long range electrostatic interactions (with the much faster pppm solver) 
         kspace_prec ... precision for the kspace solver
         cutoff ... cutoff of electrostatic interactions in Angstroms
         charge_style ... style of the interaction potential
         alpha ... alpha parameter used in some styles (like coul/dsf) given in inverse distance units
        """
        if self.lmps is None:
            lmps = lammps(cmdargs=["-screen","none"], comm=commw)
            self.lmps = lmps
        else:
            lmps = self.lmps
        lmps.command("units metal")
        if self.escut is None:
            lmps.command("atom_style charge")
            bnd_num = None
        else:
            lmps.command("atom_style hybrid charge bond")
            bnd_num = 1
        lmps.command("boundary p p p")
        lmps.command("box tilt large")
        lmps.command("atom_modify map yes")  # this is necessary for atomic styles otherwise scatter atoms does not work
        if chargefile is not None:
            charges = np.genfromtxt(chargefile).ravel()
            init_geom.set_charges(charges)
        # HACK here, we write to a file from which we read to initialize lammps - which is not exactly pretty
        init_geom.write_file("tmp.pos",filetype="lmp_charge",bond_num=bnd_num)
        
        self._read_lmp_special()
        
        
        if charge_style == "gauss/dsf":
            lmps.command("pair_style buck6d/coul/gauss/dsf     0.9000     0.9000    %12.6f" % (cutoff))
        elif charge_style == "coul/cut":
            lmps.command("pair_style coul/cut  %12.6f" % (cutoff))
        elif charge_style == "coul/dsf":   # dsf has a smooth decay to zero - so it should be superior - it also has the erf definition however maybe there is not such a large difference compared to the buck6d potential - but the alpha parameter is now global - if it is very low it should just behave like regular point charges
            lmps.command("pair_style coul/dsf  %12.6f  %12.6f" % (alpha, cutoff))
        elif charge_style == "coul/wolf":
            lmps.command("pair_style coul/wolf  %12.6f  %12.6f" % (alpha, cutoff))
        else:
            assert False, "ERROR: charge style %s is not known" % charge_style
        
            
        
        if charge_style == "gauss/dsf":
            # here need to add the pair coefficients for the rest of the potential
            # the question here is how to obtain the parameters - is it reasonable to have them in a python data file as a dictionary?
            # I cannot only use the elems here - as different atom type based potentials might be used
            if init_geom.atom_types is not None:
                syselems = init_geom.atom_types
            else:
                syselems = init_geom.elems
            uelems,uinds = np.unique(syselems,return_index=True)
            logical = np.argsort(uinds)
            uelems = uelems[logical]
            uinds = uinds[logical]  # to get the actual element
            for eid1, at1 in enumerate(uelems):
                for eid2, at2 in enumerate(uelems):
                    if eid2 >= eid1:  # interactions are symmetric
                        elem1 = init_geom.elems[uinds[eid1]]
                        elem2 = init_geom.elems[uinds[eid2]]
                        sigma1 = elems.sigmas[elem1.lower()]
                        sigma2 = elems.sigmas[elem2.lower()]
                        # mixing is not concerning in this case because this is mostly relevant for the close
                        # range interactions
                        sigma_pair = (sigma1**2 + sigma2**2)**0.5 # mixing rule
                        # no vdW - only coulomb
                        lmps.command("pair_coeff     %d     %d     0   0   0   0      %12.8f" %
                                     (eid1+1,eid2+1,sigma_pair))
        else:
            lmps.command("pair_coeff * *") # for all styles that do not have atom type dependent definitions
        
        
        self.geom = init_geom
        
        self._post_setup(lmps)

        
        self.natoms = len(self.geom.coors)
        return self.lmps 
    
    def _read_lmp_special(self):
        if self.escut is None:
            self.lmps.command("read_data tmp.pos")
        else:
            self.lmps.command("read_data tmp.pos extra/bond/per/atom %d extra/special/per/atom %d" %
                              (self.extra_bonds,self.extra_bonds*4))
    
    def _post_setup(self,lmps):
        """
          method to setup pressure and other things that are required for all supported potential types
          additon: set up exclusion range for electrostatic interactions
        """
        self.pressure_elements = ["pxx", "pyy", "pzz", "pxy", "pxz", "pyz"]
        for p in self.pressure_elements:
            lmps.command("variable %s equal %s" % (p,p))
            
        lmps.command('thermo_style custom step temp etotal pe ke pxx pyy pzz')
        
        
        # short range exclusion for electrostatic interactions:
        if self.escut is not None:
            if self.geom.atom_types is not None:
                syselems = self.geom.atom_types
            else:
                syselems = self.geom.elems
            uelems,uinds = np.unique(syselems,return_index=True)
            logical = np.argsort(uinds)
            uelems = uelems[logical]
            uinds = uinds[logical]  # to get the actual element
            fixid = 0
            for eid1, at1 in enumerate(uelems):
                for eid2, at2 in enumerate(uelems):
                    if eid2 >= eid1:  # interactions are symmetric
                        fixid += 1
                        lmps.command("fix bcreate%d all bond/create 1 %d %d %f 1" % (fixid, eid1+1, eid2+1, self.escut))
            # to update immediately, compute a single point
            self.lmps.command("run 1 pre yes post no")
            for fid in range(fixid):
                lmps.command("unfix bcreate%d" % (fid+1))
            # it might be better if the bonds cannot be broken however
            # otherwise there might be discontinuities in the PES
            #lmps.command("fix bbreak  all bond/break 100 1 %f" % (self.escut))
            # this will dynamically create/break bonds every 100 time steps within a specific cutoff
            # the problem here is probably, that this is a HARD cutoff. It would be much nicer if there
            # was a smooth transition
            lmps.command("special_bonds lj 0.00 0.00 0.00 coul 0.00 0.00 0.00")
    
    def update_coors(self, cell, coors):
        """
        update the coordinates internally in the lammps object
        """
        comm = "change_box all x final %16.10f %16.10f y final %16.10f %16.10f z final %16.10f %16.10f xy final %16.10f xz final %16.10f yz final %16.10f" % (
            0.,cell[0,0],0.,cell[1,1],0.,cell[2,2],cell[1,0],cell[2,0],cell[2,1])
        self.lmps.command(comm)
        self.lmps.scatter_atoms("x",1,3,np.ctypeslib.as_ctypes(coors))
        return
    
    def update_cell(self, cell):
        """
        update the size of the simulation box and rescale the atoms
        """
        comm = "change_box all x final %16.10f %16.10f y final %16.10f %16.10f z final %16.10f %16.10f xy final %16.10f xz final %16.10f yz final %16.10f %s" % (
            0.,cell[0,0],0.,cell[1,1],0.,cell[2,2],cell[1,0],cell[2,0],cell[2,1],
            "remap units box")
        self.lmps.command(comm)
        return
    
    def relax_int(self, thresh=1.0e-8, method="cg", etol=0.0, maxiter=1000, maxeval=1000):
        """
        relax internal coordinates only
        """
        assert method in ["cg", "hftn", "sd"]
        self.lmps.command("min_style %s" % method)
        self.lmps.command("minimize %f %f %d %d" % (etol, thresh, maxiter, maxeval))
        return
    
    def relax_box(self, thresh=1.0e-8, method="cg", etol=0.0, maxiter=100, maxeval=1000, maxstep=10, iso="tri", p=0):
        """
        relax the coordinates including the lattice vectors
        """
        assert method in ["cg", "sd"]
        stop = False
        self.lmps.command("min_style %s" % method)
        self.lmps.command("minimize %f %f %d %d" % (etol, thresh, maxiter, maxeval))
        counter = 0
        while not stop:
            self.lmps.command("fix latmin all box/relax %s %f vmax 0.01" % (iso, p))            
            self.lmps.command("minimize %f %f %d %d" % (etol, thresh, maxiter, maxeval))
            self.lmps.command("unfix latmin")
            self.lmps.command("min_style %s" % method)
            self.lmps.command("minimize %f %f %d %d" % (etol, thresh, maxiter, maxeval))
            counter += 1
            if counter >= maxstep: 
                stop=True
        return
    
    def get_xyz(self):
        """
        get the xyz position as a numpy array
        """
        self.xyz = np.ctypeslib.as_array(self.lmps.gather_atoms("x",1,3))
        self.xyz.shape=(self.natoms,3)
        return self.xyz
    
    def calc_force(self):
        """
        get the forces from the current displacement and return a numpy array
        """
        self.lmps.command("run 0 pre yes post no")
        
        self.forces = np.ctypeslib.as_array(self.lmps.gather_atoms("f",1,3))
        self.forces.shape=(self.natoms,3)
        return self.forces
    
    def calc_energy(self):
        """
        calculates and returns the total energy of a single point
        """
        self.lmps.command("run 0 pre yes post no")
        self.energy = self.lmps.get_thermo("etotal")
        return self.energy
    
    def get_force(self):
        """
        get the forces from the current displacement and return a numpy array
        """
        
        self.forces = np.ctypeslib.as_array(self.lmps.gather_atoms("f",1,3))
        self.forces.shape=(self.natoms,3)
        return self.forces
    
    def get_energy(self):
        """
        returns the total energy of a single point
        """
        self.energy = self.lmps.get_thermo("etotal")
        return self.energy
    
    def get_cell(self):
        """
        return the current cell in matrix form
        """
        var = ["boxxlo", "boxxhi", "boxylo", "boxyhi", "boxzlo", "boxzhi", "xy", "xz", "yz"]
        cell_raw = {}
        for v in var:
            cell_raw[v] = self.lmps.extract_global(v)
        # currently only orthorombic
        cell = np.zeros([3,3],"d")
        cell[0,0]= cell_raw["boxxhi"]-cell_raw["boxxlo"]
        cell[1,1]= cell_raw["boxyhi"]-cell_raw["boxylo"]
        cell[2,2]= cell_raw["boxzhi"]-cell_raw["boxzlo"]
        # set tilting
        cell[1,0] = cell_raw['xy']
        cell[2,0] = cell_raw['xz']
        cell[2,1] = cell_raw['yz']
        return cell
    
    def get_stress_tensor(self):
        """
        get the current stress tensor (in metal units - bar)
        """
        ptensor_flat = np.zeros([6])
        for i,p in enumerate(self.pressure_elements):
            ptensor_flat[i] = self.lmps.extract_variable(p, None, 0)
        ptensor = np.zeros([3,3], "d")
        # "pxx", "pyy", "pzz", "pxy", "pxz", "pyz"
        ptensor[0,0] = ptensor_flat[0]
        ptensor[1,1] = ptensor_flat[1]
        ptensor[2,2] = ptensor_flat[2]
        ptensor[0,1] = ptensor_flat[3]
        ptensor[1,0] = ptensor_flat[3]
        ptensor[0,2] = ptensor_flat[4]
        ptensor[2,0] = ptensor_flat[4]
        ptensor[1,2] = ptensor_flat[5]
        ptensor[2,1] = ptensor_flat[5] 
        return ptensor
