#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 28 14:17:03 2022

@author: sandro_w
"""

import numpy as np

import argparse
import sys
import matplotlib.pyplot as plt
import util

def get_elastic_from_OUTCAR(fname):
    elastic = None
    
    with open(fname) as fp:
        elastic = []
        read_elastic = False
        for line in fp:
            elements = line.split()
            if len(elements) > 3:
                if ((elements[0] == "TOTAL") & (elements[1] == "ELASTIC") & 
                    (elements[2] == "MODULI") & (elements[3] == "(kBar)")):
                    read_elastic = True
            if ((len(elements) == 7) & read_elastic):
                if (elements[0]!="Direction"):
                    elastic.append(list(map(float,elements[1:])))
                    if len(elastic) > 5:
                        break
        fp.close()
        elastic = np.array(elastic)/10
    return elastic


def decide_to_plot(elastic, numsys, ztol, prev_elastic):
    """
      function to check if it makes sense to plot this particular matrix element
    """
    plot_this = True
    
    # check if element is zero
    if np.allclose(elastic,np.zeros(numsys),atol=ztol):
        plot_this = False
    # check if a previously plotted entry is within the tolerance
    for pel in prev_elastic:
        if np.allclose(elastic,pel,atol=ztol):
            plot_this = False
    return plot_this
    

def plot_vasp_elastic_comparison(fnames, xlabel="dirnames", noshow=True, ztol=1e-3, sep=False,
                                 xticks=None, figsize=None, nolegend=False):
    """
      main function
    """
    
    pltstyle="~/ez-grp_style.mplstyle"
    plt.style.use(pltstyle)
    
    diagonals = [[0,0],[1,1],[2,2],[3,3],[4,4],[5,5]]
    diag_labels = ["$C_{11}$","$C_{22}$","$C_{33}$","$C_{44}$","$C_{55}$","$C_{66}$"]
    diag_markers = ["o","v","s","^","P","*"]
    diag_colors = ["C0","C1","C2","C3","C4","C5"]
    upper_diag_m = [[0,1],[0,2],[1,2]]
    u_diag_labels = ["$C_{12}$","$C_{13}$","$C_{23}$"]
    u_diag_markers = ["<",">","p"]
    u_diag_colors = ["C6","C7","C8"]
    upper_odiag_m = [[0,3],[0,4],[0,5],[1,3],[1,4],[1,5],[2,3],[2,4],[2,5],[3,4],[3,5],[4,5]]
    u_odiag_labels = ["$C_{14}$","$C_{15}$","$C_{16}$",
                      "$C_{24}$","$C_{25}$","$C_{26}$",
                      "$C_{34}$","$C_{35}$","$C_{36}$",
                      "$C_{45}$","$C_{46}$","$C_{56}$"]
    u_odiag_markers = [".","1","2","3","4","8",
                       "h","H","+","X","d","D"]
    u_odiag_colors = ["C0","C1","C2","C3","C4","C5",
                      "C6","C7","C8","C9","C7","C9"]
    
    fignames = ["C_compare_diag.png","C_compare_odiagl.png","C_compare_odiag.png"]
    
    all_elastic = []
    dirnames = []
    for fname in fnames:
        all_elastic.append(get_elastic_from_OUTCAR(fname))
        if xlabel=="dirnames":
            dirnames.append("/".join(fname.split("/")[:-1]))
    
    all_elastic = np.array(all_elastic)
    
    numsys = len(fnames)
    figures = []
    
    f1 = plt.figure(1, figsize=figsize)
    figures.append(f1)
    if sep:
        f2 = plt.figure(figsize=figsize)
        f3 = plt.figure(figsize=figsize)
        figures.append(f2)
        figures.append(f3)
        
        
    markersize = 12
    
    
    if sep:
        plt.sca(f1.gca())
    prev_elastic = []
    for did, ind in enumerate(diagonals):
        if decide_to_plot(all_elastic[:,ind[0],ind[1]],numsys,ztol,prev_elastic):
            plt.plot(range(numsys), all_elastic[:,ind[0],ind[1]],label=diag_labels[did],
                     marker=diag_markers[did],markersize=markersize,color=diag_colors[did])
            prev_elastic.append(all_elastic[:,ind[0],ind[1]])
        
    if sep:
        plt.sca(f2.gca())
    prev_elastic = []
    for did, ind in enumerate(upper_diag_m):
        if decide_to_plot(all_elastic[:,ind[0],ind[1]],numsys,ztol,prev_elastic):
            plt.plot(range(numsys), all_elastic[:,ind[0],ind[1]],label=u_diag_labels[did],
                     marker=u_diag_markers[did],markersize=markersize,color=u_diag_colors[did])
            prev_elastic.append(all_elastic[:,ind[0],ind[1]])
        
    if sep:
        plt.sca(f3.gca())
    prev_elastic = []
    for did, ind in enumerate(upper_odiag_m):
        if decide_to_plot(all_elastic[:,ind[0],ind[1]],numsys,ztol,prev_elastic):
            plt.plot(range(numsys), all_elastic[:,ind[0],ind[1]],label=u_odiag_labels[did],
                     marker=u_odiag_markers[did],markersize=markersize,color=u_odiag_colors[did])
            prev_elastic.append(all_elastic[:,ind[0],ind[1]])
    
    if figsize is not None:
        figx = figsize[0]
    else:
        figx = 6.4
    for fid,fig in enumerate(figures):
        plt.sca(fig.gca())
        if not nolegend:
            plt.legend(ncol=4,fontsize=12,bbox_to_anchor=(0.2/6.4*figx, 1.05))
        if xticks is not None:
            plt.xticks(range(numsys),xticks,rotation=45,fontweight="bold",horizontalalignment="right")
            if xlabel != "dirnames":
                plt.xlabel(xlabel)
        elif xlabel == "dirnames":
            plt.xticks(range(numsys),dirnames,rotation=45,fontweight="bold",horizontalalignment="right")
        else:
            plt.xlabel(xlabel)
        plt.ylabel("C / GPa")
        plt.tight_layout()
        plt.savefig(fignames[fid])
        
    for eid, elastic in enumerate(all_elastic):
        if eid != len(all_elastic)-1:
            el_rmsd = util.rmse(elastic.ravel(),all_elastic[-1].ravel())
            print("RMSD of entry %d compared to last: %16.10f" % (eid, el_rmsd))
            
    if not noshow:
        plt.show()
    
    
    return



if __name__ == "__main__":
  parser = argparse.ArgumentParser(description="compares different elastic moduli available in a VASP format")


  parser.add_argument("-xlabel", dest="xlabel", type=str, default="dirnames", 
                      help="set the xlabel in the plot")
  
  
  parser.add_argument("-xticks", dest="xticks", type=str, nargs="*", default=None, 
                      help="specify custom xticks")
  
  parser.add_argument("-figsize", dest="figsize", type=float, nargs=2, default=None, 
                      help="specify the size of the figures, provide x and y values")
  
  parser.add_argument("-ztol", dest="ztol", type=float, default=1e-3, 
                      help="zero tolerance - matrix elements below this value will not be visualized")
  
  parser.add_argument("-noshow", dest="noshow", action="store_true", default=False, 
                      help="do not show plots")
  
  parser.add_argument("-nolegend", dest="nolegend", action="store_true", default=False, 
                      help="do not show the legend")
  
  parser.add_argument("-sep", dest="sep", action="store_true", default=False, 
                      help="3 separate plots in multiple containing diagonals, off-diagonals of 1,2,3 and the rest")
  
  parser.add_argument("fnames", nargs = "*", help="paths to all the outcars that should be compared")


  args = parser.parse_args()
  
  afp = open("p_cmnd.log","a")
  afp.write(" ".join(sys.argv)+"\n")
  afp.close()
  
  
  
  plot_vasp_elastic_comparison(**vars(args))