#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
  this script serves to evaluate an AEMD run and compute the thermal diffusivity
  from the appropriate script
"""

import sys
import argparse
import numpy as np
import subprocess
from scipy import optimize
import util
import matplotlib.pyplot as plt
import geometry
from swfileIO import dump

# function to be fit
# assuming T2 is the higher temperature
# taken from: Melis, C., Dettori, R., Vandermeulen, S., & Colombo, L. (2014). Calculating thermal conductivity in a transient conduction regime: theory and implementation. European Physical Journal B, 87(4), 1–9. https://doi.org/10.1140/epjb/e2014-50119-0
def delta_T_func(t, diffusivity, init_T1, init_T2, Lz, order):
    delta_T = 0
    for oid in range(order):
        c_order = oid+1 # start counting at 1
        alpha_n = 2*np.pi*c_order/Lz
        Cn = 8*(init_T2-init_T1)*(np.cos(alpha_n*Lz/2)-1)**2/(alpha_n*Lz)**2
        delta_T += Cn*np.exp(-alpha_n**2*diffusivity*t)
    return delta_T

# use subprocess to execute a command and return the contents
def exec_cmd(cmd):
    p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    strings, err = p.communicate()
    return strings.decode().split()

# function to read Cv from file fname at temperature temp
def read_tprops(fname, temp):
    Cv = None
    with open(fname) as fp:
        readcv = False
        for line in fp:
            elements = line.split()
            if len(elements) > 1:
                if elements[1] == "temperature:":
                    if abs(float(elements[2])-temp) < 1e-3:
                        readcv = True
                elif (elements[0] == "heat_capacity:") & readcv:
                    Cv = float(elements[1])
                    break
        fp.close()
    return Cv


def eval_aemd(noshow=False, units="metal", order=3, dir="x", init_guess=0.2, cv=None, phstruct="POSCAR",
              tmax=None, profplot=False, ravgwidth=1.5, tconv=False, ntconv=20, oconv=False, omax=20,
              cv_dp=False):
    
    plt.style.use("~/ez-grp_style.mplstyle")
    
    # need to get the time step
    timestep = float(exec_cmd("grep timestep AEMD.in".split())[3])
    
    print("timestep:",timestep)
    
    # need to also get hot and cold thermostat temperature at the beginning
    itemp_cold = float(exec_cmd("grep temp_cold AEMD.in".split())[3])
    itemp_hot  = float(exec_cmd("grep temp_hot AEMD.in".split())[3])
    
    print("cold initial temperature:",itemp_cold,"K")
    print("hot initial temperature:",itemp_hot,"K")
    
    # read the actual average temperatures
    dtemp = np.genfromtxt("temp_diff.dat")
    dt    = dtemp[:,0]
    tcold = dtemp[:,1]
    thot  = dtemp[:,2]
    
    direc = util.dirmap(dir)
    
    # need to obtain the length from  the dump file
    xyz_latparams = exec_cmd("head dump_vels.dat -n 8".split())
    if len(xyz_latparams) > 24:
        # for triclinic cells
        Lz = float(xyz_latparams[-(2-direc)*3-2])-float(xyz_latparams[-(2-direc)*3-3])   # xyzhi - xyzlo
    else:
        # this works for a othorhombic cell
        Lz = float(xyz_latparams[-(2-direc)*2-1])-float(xyz_latparams[-(2-direc)*2-2])   # xyzhi - xyzlo
    
    print("cell length in A:",Lz)
    
    
    # convert quantities entering the equation in SI units
    if units == "metal":
        # convert time in picoseconds
        dt *= timestep*1e-12
        Lz *= 1e-10
    elif units == "real":
        # convert time in femtoseconds
        timestep *= 1e-3
        dt *= timestep*1e-12
        Lz *= 1e-10
    else:
        print("ERROR: unit type not known, type metal or real are supported")
        return
    
    
    print("performing an order %d fit..." % order)
    # build the fitfunction
    def fitfunc(t, diffusivity):
        return delta_T_func(t, diffusivity, itemp_cold, itemp_hot, Lz, order)
    
    
    delta_T_sim = thot-tcold
    
    # initial guess for the diffusivity
    # convert in SI
    print("initial guess for the thermal diffusivity:",init_guess,"mm^2/s")
    guesses = [init_guess/1e6]
    
    if tmax is not None:
        print("evaluating the first %d time steps" % tmax)
        tvals = dt[:tmax]
        DTvals = delta_T_sim[:tmax]
        plt.axvline(tmax*timestep, color="r")
    else:
        tvals = dt
        DTvals = delta_T_sim
    
    # perform the actual fit
    params, cov = optimize.curve_fit(fitfunc, tvals, DTvals, p0=guesses)
    
    diffusivity = params[0]
    
    print("covariance matrix:")
    print(cov)
    
    # 1 is 1 second
    print("temperature difference in the long time limit (this should be zero):", fitfunc(1, diffusivity),"K")
    
    print("thermal diffusivity:", diffusivity, "m^2/s")
    print("thermal diffusivity:", diffusivity*1e6, "mm^2/s")
    
    plt.plot(dt*1e12, delta_T_sim, label="raw data")
    plt.plot(dt*1e12, fitfunc(dt, diffusivity),"k-", label="fit")
    ravg, nvals = util.running_avg(delta_T_sim, 500)
    ravgx = np.linspace(min(dt)*1e12,max(dt)*1e12,nvals)
    plt.plot(ravgx, ravg, "k", label="ravg", alpha=0.25)
    plt.axhline(0,linestyle="--",alpha=0.5,color="k")
    
    plt.xlabel("t / ps")
        
    plt.ylabel("$\Delta T$ / K")
    
    plt.legend()
    
    plt.tight_layout()
    
    plt.savefig("diffusivity_fit.png")
    
    
    Na = 6.02214076*10**23
    kB = 1.380649*10**-23
    if cv is not None:
        # get cv
        meantemp = (itemp_hot+itemp_cold)/2
        heat_capacity = read_tprops(cv, meantemp)
        print("read cv at temperature %6.1f K: %f J/K/mol" % (meantemp, heat_capacity))
        print("assuming the base unit cell of the phonon calculation is the primitive unit cell (no primitive axis)")
        heat_capacity /= Na  # get rid of the mol
        
        # get the volume
        print("reading volume from file:",phstruct)
        geo = geometry.geo(phstruct, filetype="POSCAR")
        volume = geo.get_volume()
        print("volume:",volume,"A^3")
        kappa = diffusivity/volume/1e-30*heat_capacity
        print("thermal conductivity:",kappa,"W/mK")
    
    # dulong petit limit
    if cv_dp:
        print("thermal conductivity in the Dulong-Petit limit:")
        lmp_dump = dump.dump("dump_vels.dat", filetype="fxyz", readmax=1)
        cell = lmp_dump.cells[0]
        volume = np.dot(cell[0],np.cross(cell[1],cell[2]))
        numats = int(xyz_latparams[7])
        heat_capacity = 3*kB*numats
        print("heat capacity: ",heat_capacity,"J/K")
        print("volume: ",volume,"A^3")
        kappa = diffusivity/volume/1e-30*heat_capacity
        print("thermal conductivity: ",kappa,"W/mK")
        
        
        
    
    # to visualize the temperature evolution
    if profplot:
        print("reading trajectory from dump_vels.dat...")
        lmp_dump = dump.dump("dump_vels.dat", filetype="fxyz")
        velocities = lmp_dump.forces  # these are actually not forces
        # velocities are given in A/fs or A/ps depending on the unit
        norm_velocities = np.linalg.norm(velocities,axis=2)
        # converting to m/s
        if units == "metal":
            norm_velocities *= 10**2
        elif units == "real":
            norm_velocities *= 10**5
        dumpcoor = lmp_dump.coors
        dumptime = lmp_dump.timestep * timestep
        masses = lmp_dump.get_atom_masses()*1.660539*10**-27  # masses in amu - need to convert to kg
        
        
        plt.figure(2)
        
        vel_inds_to_plot = [0,5,10,15,20,25]
        
        for eid in vel_inds_to_plot:
            atemp = masses*norm_velocities[eid]**2/3/kB # atomic "temperature"
            # here, we essentially need to plot something like a running average - atomic temperatures are nonsensical
            xdat = np.linspace(0,Lz/1e-10,200)
            ydat = np.zeros(np.shape(xdat))
            dircoor = dumpcoor[eid,:,direc]
            for xid, xd in enumerate(xdat):
                logical = np.abs(dircoor-xd) < ravgwidth
                ydat[xid] = np.mean(atemp[logical])
                
            plt.plot(xdat,ydat,label=str(dumptime[eid]))
        plt.axhline(itemp_cold,linestyle="--",alpha=0.5,color="k")
        plt.axhline(itemp_hot,linestyle="--",alpha=0.5,color="k")
        if len(vel_inds_to_plot) < 10:
            plt.legend(fontsize=10)
        plt.xlabel("L / $\AA$")
        plt.ylabel("T / K")
        plt.tight_layout()
        
        plt.savefig("temp_prof.png")
    
    if tconv:
        
        print("evaluating time convergence...")
        t_test = np.linspace(len(dt)/10,len(dt),ntconv)
        
        test_times = []
        diffusivities = []
        for entry in t_test:
            tvals = dt[:round(entry)]
            DTvals = delta_T_sim[:round(entry)]
            
            # perform the actual fit
            params, cov = optimize.curve_fit(fitfunc, tvals, DTvals, p0=guesses)
            test_times.append(round(entry)*timestep)
            diffusivities.append(params[0]*1e6)
            
        
        plt.figure(3)
        plt.plot(test_times, diffusivities)
        
        plt.xlabel("t / ps")
        plt.ylabel("diffusivity / $mm^2/s$")
        plt.tight_layout()
        plt.savefig("aemd_tconv.png")
    
    if oconv:
        
        print("evaluating series order convergence...")
        o_test = np.array(range(omax))+1
        
        diffusivities = []
        for entry in o_test:
            
            # perform the actual fit
            def fitfunc_conv(t, diffusivity):
                return delta_T_func(t, diffusivity, itemp_cold, itemp_hot, Lz, entry)
            params, cov = optimize.curve_fit(fitfunc_conv, tvals, DTvals, p0=guesses)
            diffusivities.append(params[0]*1e6)
            
        
        plt.figure(4)
        plt.plot(o_test, diffusivities)
        
        plt.xlabel("series order")
        plt.ylabel("diffusivity / $mm^2/s$")
        plt.tight_layout()
        plt.savefig("aemd_oconv.png")
    
    if not noshow:
        plt.show()
        
        
    return diffusivity, kappa


if __name__ == "__main__":
    
    parser = argparse.ArgumentParser(description="This script serves to actually obtain the thermal diffusivity from an AEMD simulation")
    
    
    # arguments for MD evaluation
    parser.add_argument("-noshow", dest="noshow", action="store_true", default=False, help="do not show plots")
    parser.add_argument("-units", dest="units", type=str, default="metal", help="lammps unit style")
    parser.add_argument("-order", dest="order", type=int, default=3, help="order of the diffusivity fit")
    parser.add_argument("-guess", dest="init_guess", type=float, default=0.2, help="initial guess for the thermal diffusivity in mm^2/s")
    parser.add_argument("-dir", dest="dir", type=str, default="x", help="direction of heat transport evaluation. Be careful with non-orthorhombic cells. It might not work properly.")
    parser.add_argument("-tmax", dest="tmax", type=int, default=None, help="impose a maximum time in which the fit will be performed")
    
    # for closer evaluation
    parser.add_argument("-profplot", dest="profplot", action="store_true", default=False, help="plot the temperature profile")
    parser.add_argument("-ravgwidth", dest="ravgwidth", type=float, default=1.5, help="running average width for the temperature profile plot")
    # automatic time convergence
    parser.add_argument("-tconv", dest="tconv", action="store_true", default=False, help="plot the time convergence")
    parser.add_argument("-ntconv", dest="ntconv", type=int, default=20, help="how many tests are done to converge the time")
    
    # automatic order convergence
    parser.add_argument("-oconv", dest="oconv", action="store_true", default=False, help="plot the order convergence")
    parser.add_argument("-omax", dest="omax", type=int, default=20, help="maximum order to perform the convergence test")
    
    # arguments to obtain the thermal conductivity
    # based on the equation: kappa = diffusivity * rho * cv
    parser.add_argument("-cv", dest="cv", type=str, default=None, help="path to the thermal_properties.yaml file containing the heat capacity at the mean temperature of the initial temperatures")
    parser.add_argument("-phstruct", dest="phstruct", type=str, default="POSCAR", help="path to the POSCAR serving as input unit cell of the phonon calculation")
    parser.add_argument("-cv_dp", dest="cv_dp", action="store_true", default=False, help="use the heat capacity in the dulong-petit limit to obtain the thermal conductivity")


    args = parser.parse_args()
    
    afp = open("p_cmnd.log","a")
    afp.write(" ".join(sys.argv)+"\n")
    afp.close()
    
    eval_aemd(**vars(args))