
import numpy as np
import matplotlib.pyplot as plt
import util
import geometry
import subprocess
import scipy.signal
import scipy.interpolate
from scipy import stats
from scipy import optimize
import smooth
from sklearn.datasets import make_friedman2
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import DotProduct, WhiteKernel
from sklearn.gaussian_process.kernels import RBF, ConstantKernel as C


# these functions serve fitting purposes
# these functions assume zero is the middle of one segment
def nointerfacefunc(x,rn,k1,k2):
  y = np.zeros(np.shape(x))
  l1 = np.abs(x)<(rn/2.0)
  l2 = x>=(rn/2.0)
  l3 = x<=(-rn/2.0)
  y[l1] = k1*x[l1]
  #totlen = max(x)-min(x)
  #outlen = (totlen-rn)/2
  #middist = rn/2+outlen/2
  y[l2] = k2*(x[l2]-rn/2)+k1*rn/2
  y[l3] = k2*(x[l3]+rn/2)-k1*rn/2
  return y

def interfacefunc(x,rn,k1,k2,dint):
  y = np.zeros(np.shape(x))
  l1 = np.abs(x)<(rn/2.0)
  l2 = x>=(rn/2.0)
  l3 = x<=(-rn/2.0)
  y[l1] = k1*x[l1]
  y[l2] = k2*(x[l2]-rn/2) + dint
  y[l3] = k2*(x[l3]+rn/2) - dint
  return y

def nointerfacefunc_fixed_len(rn):
  def inner_nfunc(x,k1,k2):
    y = np.zeros(np.shape(x))
    l1 = np.abs(x)<(rn/2.0)
    l2 = x>=(rn/2.0)
    l3 = x<=(-rn/2.0)
    y[l1] = k1*x[l1]
    y[l2] = k2*(x[l2]-rn/2)+k1*rn/2
    y[l3] = k2*(x[l3]+rn/2)-k1*rn/2
    return y
  return inner_nfunc

def interfacefunc_fixed_len(rn):
  def inner_ifunc(x,k1,k2,dint):
    y = np.zeros(np.shape(x))
    l1 = np.abs(x)<(rn/2.0)
    l2 = x>=(rn/2.0)
    l3 = x<=(-rn/2.0)
    y[l1] = k1*x[l1]
    y[l2] = k2*(x[l2]-rn/2) + dint
    y[l3] = k2*(x[l3]+rn/2) - dint
    return y
  return inner_ifunc
  


class nemd_result():
  
  
  
  def __init__(self,kinavgname="",cfluxname="",refname="",profname="",units="real",
               direc=0,active_flux=False,use_lmp_atypes=False, outfile="output.lammps",
               ehex=False):
    kb_ev         = 8.6172e-5                      # Boltzman constant eV/K/atom
    kb_kc_k_mol   = 0.0019872041                   # kcal/K/mol
    kCal2J     = 4186.0/6.02214e23
    eV2J          = 1.60218e-19
    self.units_grad = 1.e-10  
    
    self.usedflux = None
  
  
    if units == "real":
      self.kb = kb_kc_k_mol
      self.units_flux = kCal2J/1e-15/(1e-10**2)  # kcla/fs/A^2 -> J/s/m^2 = W/m^2
      self.time_factor = 1
    elif units == "metal":
      self.kb = kb_ev
      self.units_flux = eV2J/1e-12/(1e-10**2)
      self.time_factor = 1000
    
    self.direc = direc

    
    if kinavgname != "":
      self.read_kins(kinavgname)
      
    if refname != "":
      self.read_ref(refname,use_lmp_atypes)
      
    if profname != "":
      self.read_tmp_profile(profname,direc)
      
    self.mpflux = False
    if cfluxname != "":
      if cfluxname == 'oflux.dat':
          self.oflux = True
      elif "mp_flux.dat" in cfluxname:
          self.oflux = False
          self.mpflux = True
      else:
          self.oflux = False
      self.read_flux(cfluxname,active_flux=active_flux,oflux=self.oflux,mpflux=self.mpflux)
    
    if ehex:
        self.flux_from_ehex(outfile)
        self.oflux = False
      
    self.oneDlens       = [[],[],[]]
    self.oneDels        = [[],[],[]]
    self.oneDtemps      = [[],[],[]]
    self.oneDlenerrs    = [[],[],[]]
    self.oneDtemperrs   = [[],[],[]]
    
    self.defaultcols = {}
    self.defaultcols["Zn"] = "darkgoldenrod"
    self.defaultcols["Ca"] = "darkolivegreen"
    self.defaultcols["O"]  = "r"
    self.defaultcols["C"]  = "k"
    self.defaultcols["H"]  = "grey"
    self.defaultcols["N"]  = "b"
    self.defaultcols["Mg"]  = [0.5,0.0,0.5]
    self.defaultcols["1"]  = "k"
    self.defaultcols["2"]  = "r"
    self.defaultcols["3"]  = "b"
    self.defaultcols["4"]  = "g"
    
    return
  
  
  
  
  # reads the last values of the heat flux from a file and averages over both directions
  # active_flux   ... calculate the real flux difference in a interval instead of the full cumulative value, only works with flux.dat
  # avgnum        ... required for active_flux, average over this many flux values
  def read_flux(self,cfluxname,active_flux=False,avgnum=5000,oflux=False,mpflux=False):
    dataflux     = np.genfromtxt(cfluxname)
    # for being able to read cflux.dat and flux.dat
    
    coloffset = np.shape(dataflux)[1] - 5
    if not oflux:
        dt = (dataflux[1,coloffset+0] - dataflux[0,coloffset+0])/(dataflux[1,0] - dataflux[0,0])
    else:
        dt = 0.5
    dt *= self.time_factor
    print ("Timestep: "+str(dt)+" fs")
    if oflux:
        self.flux1 = dataflux[:,1+self.direc]*self.units_flux
        self.flux2 = -dataflux[:,4+self.direc]*self.units_flux
        self.time_flux = dataflux[:,0]*dt
    elif mpflux:
        cross_area = self.geo.cell[(self.direc+1)%3][(self.direc+1)%3]*self.geo.cell[(self.direc+2)%3][(self.direc+2)%3]
        self.time_flux = dataflux[:,0]*dt
#        timediff = np.asarray([y - x for x, y in zip(self.time_flux, self.time_flux[avgnum:])])
#        self.time_flux = self.time_flux[avgnum:]*dt
        self.flux1 = dataflux[:,2]/cross_area/self.time_flux*self.units_flux/2*self.time_factor
        self.flux2 = dataflux[:,2]/cross_area/self.time_flux*self.units_flux/2*self.time_factor
        # the reason why the time is overwritten here is to allow for restart
        # the way lammps works the evaluation of the flux has to happen BEFORE that
        self.time_flux = np.array(range(len(self.flux1)))*dt*(dataflux[1,0] - dataflux[0,0])
    elif not active_flux:
      self.flux1        =  dataflux[:,coloffset+3]*self.units_flux
      self.flux2        =  -dataflux[:,coloffset+4]*self.units_flux
      #self.time_flux = dataflux[:,coloffset+0]   # time given in fs for real and ps for metal  
      self.time_flux = dataflux[:,0]*dt
    else:
      cross_area = self.geo.cell[(self.direc+1)%3][(self.direc+1)%3]*self.geo.cell[(self.direc+2)%3][(self.direc+2)%3]
      #print("cross_area",cross_area)
      # read cumulative energies
      Ecum1 = dataflux[:,coloffset+1]
      Ecum2 = dataflux[:,coloffset+2]
      # substract them from previous values
      E1 = np.asarray([y - x for x, y in zip(Ecum1, Ecum1[avgnum:])])
      E2 = np.asarray([y - x for x, y in zip(Ecum2, Ecum2[avgnum:])])
      self.time_flux = dataflux[:,0]   # time given in fs for real and ps for metal, only works with flux.dat!
      timediff = np.asarray([y - x for x, y in zip(self.time_flux, self.time_flux[avgnum:])])
      #print timediff
      self.time_flux = self.time_flux[avgnum:]*dt   # just so, that the shape of fluxes and timeflux is consistent and to use proper time step
      self.flux1 = E1/cross_area/timediff/dt*self.units_flux/2*self.time_factor
      self.flux2 = -E2/cross_area/timediff/dt*self.units_flux/2*self.time_factor
      
      
    self.fluxt   =  (self.flux1+self.flux2)/2.0
    self.fluxavg = (np.abs(self.flux1[-1])+np.abs(self.flux2[-1]))/2
    self.fluxerr = np.std([np.abs(self.flux1[-1]),np.abs(self.flux2[-1])])/2.0**0.5
    self.dt = dt
    return
  

    
  def flux_from_ehex(self, outfile):
      
      cmd = ["grep", "ehex\|timestep", outfile]
      p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      lines, err = p.communicate()
      lines = lines.decode().split('\n')
      
      nevery = []
      amount = []
      
      for line in lines:
          elements = line.split("#")[0].split()
          if len(elements) > 5:
              if (elements[3] == "ehex") & util.is_numeric(elements[4]) & util.is_numeric(elements[5]):
                  nevery.append(int(elements[4]))
                  amount.append(float(elements[5]))
          if len(elements) > 1:
              if (elements[0] == "timestep") & util.is_numeric(elements[1]):
                  timestep = float(elements[1])
      
        
      assert len(nevery) == 2, "ERROR: can only process two ehex entries; entries: " + str(nevery) + " " + str(amount)
      
      if amount[0]/nevery[0] != -amount[1]/nevery[1]:
          print("WARNING: positive and negative heat flux are not the same!")
          print("first:  %12.8f" % amount[0]/nevery[0])
          print("second: %12.8f" % amount[1]/nevery[1])
      
      num_tstep      = self.tproftime[-1]
      self.dt        = timestep * self.time_factor
      self.time_flux = self.dt*np.array(range(num_tstep))
      self.flux1     = np.abs(np.array([amount[0]/nevery[0] * self.units_flux] * num_tstep))
      self.flux2     = np.abs(np.array([amount[1]/nevery[1] * self.units_flux] * num_tstep))
      
      self.fluxt   =  (self.flux1+self.flux2)/2.0
      self.fluxavg = (np.abs(self.flux1[-1])+np.abs(self.flux2[-1]))/2
      self.fluxerr = np.std([np.abs(self.flux1[-1]),np.abs(self.flux2[-1])])/2.0**0.5
      
      return
                  

  # reads kinetic energies and coordinates
  def read_kins(self,kinavgname):
    data = np.genfromtxt(kinavgname, dtype=["f8","S2","f8","f8","f8","f8","f8"],
                     names=["ids","elems","x","y","z","kine","temp_atom"])
    self.temps = data["kine"]/1.5/self.kb
    self.coors = np.asarray([data["x"],data["y"],data["z"]])
    self.elems = data["elems"]
    return

  # sets the elements manually if needed
  def set_elems(self, elems):
      self.elems = elems
      self.geo.elems = elems
      self.systemels = np.unique(elems)
  
  # write the coordinates in a xyz file
  def write_coors(self,fname):
    ofp = open(fname,"w")
    ofp.write("%d\n"%(len(self.coors.T)))
    # here should be the cell in an exyz
    ofp.write("\n")
    for cid,coor in enumerate(self.coors.T):
      ofp.write("%s %16.8f %16.8f %16.8f\n" % (self.elems[cid],coor[0],coor[1],coor[2]))
    return
  
  # reads a reference geometry file to evaluate coordinates and element types
  def read_ref(self,refname,use_lmp_atypes):
    self.geo = geometry.geo(refname)
    if use_lmp_atypes:
      self.geo.elems = np.asarray(list(map(str,self.geo.lmp_species)))
    self.systemels = np.unique(self.geo.elems)
    print(self.geo.elems)
    self.natoms    = len(self.geo.elems)
    return
  
  # reads temperature profile with normal slab evaluation
  def read_tmp_profile(self,profname,direc):
    fp = open(profname,"r")
    # time dependent temperature profile
    self.tprof_t = []
    self.tproftime = []
    self.tprof_l  = []
    for line in fp:
      precomm = line.split("#")[0]
      elements = precomm.split()
      if len(elements) == 3:
        self.tproftime.append(int(elements[0]))
        self.tprof_t.append([])
        self.tprof_l = []
      if len(elements) == 4:
        self.tprof_t[-1].append(float(elements[3])/self.kb)
        self.tprof_l.append(float(elements[1])*self.geo.cell[direc][direc])
    self.tprof_l = np.asarray(self.tprof_l)
    self.tprof_t = np.asarray(self.tprof_t)
    return
  
  # evaluates temperature profile from slab evaluation
  def eval_temp_profile(self,n_convtest=100,convplots=1,fitdists=[[2.0/16,6.0/16],[10.0/16,14.0/16]],
          lowflux=False,printkappas=[1,5,10],useflux=90,figname="kappa_eval.png",noshow=True,fontsize=20,
          lnoise=True, writemisc=True, nolinfitplot=False):
    f = plt.figure(1,figsize=(12,8))
    factor_dist = 1.0/len(self.tprof_t)*max(self.time_flux/1e6)
    plt.subplot(221)

    fluxdists = np.asarray(list(map(int,np.floor(np.linspace(0,len(self.fluxt)-1,n_convtest)))))
    favg = np.zeros(np.shape(fluxdists))
    # option for large noise levels
    if lowflux:
        if self.flux1[-1]<self.flux2[-1]:
            timeflux = self.flux1
        else:
            timeflux = self.flux2
    else:
        timeflux = self.fluxt
    factor_flux = 1.0/1e6*max(self.time_flux)/len(timeflux)
    if lnoise:
      cavg = np.zeros(np.shape(fluxdists))
      havg = np.zeros(np.shape(fluxdists))
      xflvals = fluxdists*factor_flux
    for i,fdist in enumerate(fluxdists):
      favg[i] = np.mean(timeflux[fdist:])
      if lnoise:
        cavg[i] = np.mean(self.flux1[fdist:])
        havg[i] = np.mean(self.flux2[fdist:])
    if not lnoise:
        cavg = self.flux1
        havg = self.flux2
        xflvals = self.time_flux/1e6
    if not lnoise:
      plt.plot(self.time_flux/1e6,self.fluxt,label="average")
    plt.plot(xflvals,cavg/1e9,label="cold")
    plt.plot(xflvals,havg/1e9,label="hot")
    plt.vlines([fluxdists[useflux]*factor_flux],favg[useflux]*0.9,favg[useflux]*1.1)
    plt.plot(fluxdists*factor_flux,favg/1e9,label="average to end")
    if not self.oflux:
      plt.ylim([favg[useflux]/1e9*0.9,favg[useflux]/1e9*1.1])
    plt.xlabel("t / ns",fontsize=fontsize, fontweight="bold")
    plt.ylabel("J / GW/m$\mathbf{^2}$",fontsize=fontsize, fontweight="bold")
    plt.tick_params(axis='both', which='major', labelsize=fontsize-5)
    plt.grid()
    #plt.legend(fontsize=fontsize-5)
    self.fluxerr = np.std([np.abs(favg[useflux]),np.abs(favg[useflux])])/2.0**0.5 
    
    plt.subplot(222)
    dists = list(map(int,np.floor(np.linspace(0,len(self.tprof_t)-1,n_convtest))))
    kappas = np.zeros(np.shape(dists))
    kappaerrs = np.zeros(np.shape(dists))
    avgtemps = []
    all_tgrads = np.zeros(np.shape(dists))
    all_tgraderrs = np.zeros(np.shape(dists))
    plotlist = list(map(int,np.linspace(0,len(dists)-1,convplots)))
    if convplots == 1:
        plotlist = [printkappas[-1]]
    prevdist = 0
    fitdists = np.asarray(fitdists) * max(self.tprof_l)
    for i,dist in enumerate(dists):
      tprof_avg = np.mean(self.tprof_t[dist:,:],axis=0)
      Tgrads = []
      for interval in fitdists:
        logical = (self.tprof_l > interval[0]) & (self.tprof_l < interval[1])
        fxvals = self.tprof_l[logical]
        fyvals = tprof_avg[logical]
        fit = np.polyfit(fxvals,fyvals,1)
        fit_fn = np.poly1d(fit)
        Tgrads.append(abs(fit_fn[1])/self.units_grad)
        if i in plotlist:
          ccol = "C"+str(plotlist.index(i))
          if not nolinfitplot:
              plt.plot([min(fxvals),max(fxvals)],fit_fn([min(fxvals),max(fxvals)]),zorder=3+(len(plotlist)-plotlist.index(i)),color="k",linewidth = 6)
      if i in plotlist:
        ccol = "C"+str(plotlist.index(i))
        plt.plot(self.tprof_l,tprof_avg,"o",label=str(float(dist)*factor_dist),color=ccol,zorder=len(plotlist)-plotlist.index(i))
        np.savetxt("tprof.dat",np.asarray([self.tprof_l,tprof_avg]).T)
      # evaluate temperature gradient and therefore thermal conductivity
      Tgrad = np.mean(Tgrads)
      Tgraderr = np.std(Tgrads)/len(Tgrads)**0.5
      avgtemps.append(tprof_avg)
      all_tgrads[i] = Tgrad
      all_tgraderrs[i] = Tgraderr
      kappas[i] = favg[useflux] / Tgrad
      kappaerrs[i] = favg[useflux] / Tgrad**2 * Tgraderr + self.fluxerr / Tgrad
      
    
    plt.xlabel("L / $\mathbf{\AA}$",fontsize=fontsize, fontweight="bold")
    plt.ylabel("T / K",fontsize=fontsize, fontweight="bold")
    if convplots > 1:
      plt.legend()
    plt.tick_params(axis='both', which='major', labelsize=fontsize-5)
    plt.grid()
    
    
    plt.subplot(212)
    fdist = np.asarray(list(map(float,dists)))*factor_dist
    plt.plot(fdist,kappas)
    plt.ylabel("$\mathbf{\kappa}$ / W/mK",fontsize=fontsize, fontweight="bold")
    plt.xlabel("t / ns",fontsize=fontsize, fontweight="bold")
    np.savetxt("kappaevol.dat",np.asarray([fdist,kappas]).T)
    plt.tick_params(axis='both', which='major', labelsize=fontsize-5)
    for vis in printkappas:
      plt.vlines([dists[vis]*factor_dist],min(kappas)*0.9,max(kappas)*1.1)
      print ("Thermal conductivity %12.6f: %12.6f  %12.6f (error) W/mK" % (dists[vis]*factor_dist,kappas[vis],kappaerrs[vis])) 
    plt.grid()
    # writes several additional values in an output file
    if writemisc:
      ofp = open("tslab_result_misc.dat","w")
      self.slabevaldicts = []
      for vis in printkappas:
        p = subprocess.Popen(["grep", "variable dbin", "NEMD.in"], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        results, err = p.communicate()
        results = results.decode().split()
        if "numbin" in results[3]:
            p = subprocess.Popen(["grep", "variable numbin", "NEMD.in"], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            results, err = p.communicate()
            results = results.decode().split()
            number_of_slabs = int(results[3])
        else:
            number_of_slabs = int(results[3])-1
        miscdict = {}
        ofp.write("Thermal conductivity at %12.6f Ang:\n %12.6f # kappa [W/mK]\n %12.6f # error_kappa [W/mK]\n" % (dists[vis]*factor_dist,kappas[vis],kappaerrs[vis]))
        ofp.write(" %12.6f # averaged heat flux used for evaluation [W/m^2] \n" % (favg[useflux]))
        ofp.write(" %12.6f # temperature gradient [K/m]\n" % (all_tgrads[vis]))
        ofp.write(" %12.6f # temperature gradient error [K/m]\n" % (all_tgraderrs[vis]))
        ofp.write(" %12.6f # first slab position [Ang]\n" % (np.min(self.tprof_l)))
        ofp.write(" %12.6f # last slab position [Ang]\n" % (np.max(self.tprof_l)))
        ofp.write(" %12.6f # min temperature [K]\n" % (np.min(avgtemps[vis])))
        mean_temp = np.mean(avgtemps[vis])
        ofp.write(" %12.6f # average temperature [K]\n" % (mean_temp))
        ofp.write(" %12.6f # max temperature [K]\n" % (np.max(avgtemps[vis])))
        cell_len = self.geo.cell[self.direc][self.direc]
        # this section has to be adjusted for MP - there the slabs are distributed differently
        slablength = cell_len/(number_of_slabs)
        if self.mpflux:
            tstat_cold_1 = slablength
            tstat_hot_1 = cell_len/2
            tstat_hot_2 = cell_len/2 + slablength
            tstat_cold_2 = cell_len
            tstat_dist = tstat_hot_1-tstat_cold_1
        else:
            tstat_cold_1 = slablength/2
            tstat_hot_1 = cell_len/2 - slablength/2
            tstat_hot_2 = cell_len/2 + slablength/2
            tstat_cold_2 = cell_len - slablength/2
            tstat_dist = tstat_hot_1-tstat_cold_1
        ofp.write(" %12.6f # cell length [Ang]\n" % (cell_len))
        zero_intercept = (-tstat_dist*self.units_grad/2)*all_tgrads[vis]+mean_temp
        ofp.write(" %12.6f # zero intercept [K]\n" % (zero_intercept))
        half_intercept = ( tstat_dist*self.units_grad/2)*all_tgrads[vis]+mean_temp
        ofp.write(" %12.6f # half intercept [K]\n" % (half_intercept))
        ofp.write(" %12.6f # intercept difference [K]\n" % (half_intercept-zero_intercept))
        ofp.write(" %12.6f # 100 - intercept difference/2 [K]\n" % ((100-(half_intercept-zero_intercept))/2))
        tprof_avg = avgtemps[vis]
        outer_slab_temps = tprof_avg[(self.tprof_l < tstat_cold_1) | (self.tprof_l > tstat_cold_2)]
        outer_slab_temps = outer_slab_temps[outer_slab_temps>5]
        inner_slab_temps = tprof_avg[(self.tprof_l > tstat_hot_1) & (self.tprof_l < tstat_hot_2)]
        inner_slab_temps = inner_slab_temps[inner_slab_temps>5]
        outer_tstat_temp = np.mean(outer_slab_temps)
        inner_tstat_temp = np.mean(inner_slab_temps)
        ofp.write(" %12.6f # outer thermostat temperature\n" % (outer_tstat_temp))
        ofp.write(" %12.6f # inner thermostat temperature\n" % (inner_tstat_temp))
        ofp.write(" %12.6f # thermostat temperature difference\n" % (inner_tstat_temp-outer_tstat_temp))
        miscdict["kappa"] = kappas[vis]
        miscdict["kappaerrs"] = kappaerrs[vis]
        miscdict["flux"] = favg[useflux]
        miscdict["Tgrad"] = all_tgrads[vis]
        miscdict["Tgraderr"] = all_tgraderrs[vis]
        miscdict["cell_len"] = cell_len
        miscdict["half_intercept"] = half_intercept
        miscdict["zero_intercept"] = zero_intercept
        miscdict["tstat_diff"] = half_intercept-zero_intercept
        plt.subplot(222)
        xtslpvals = np.linspace(tstat_cold_1,tstat_hot_1,10)-tstat_cold_1
        ytslpvals = (xtslpvals-tstat_dist/2)*self.units_grad*all_tgrads[vis]+mean_temp
        if not nolinfitplot:
            plt.plot(xtslpvals+tstat_cold_1,ytslpvals,linestyle="dotted",color="k")
        ytslpvals = (xtslpvals-tstat_dist/2)*-self.units_grad*all_tgrads[vis]+mean_temp
        if not nolinfitplot:
            plt.plot(xtslpvals+tstat_hot_2,ytslpvals,linestyle="dotted",color="k")
#        plt.vlines([tstat_cold_1,tstat_cold_2],mean_temp-50,mean_temp+50,color="b")
#        plt.vlines([tstat_hot_1,tstat_hot_2],mean_temp-50,mean_temp+50,color="r")
        plt.axvline(tstat_cold_1,color="b")
        plt.axvline(tstat_cold_2,color="b")
        plt.axvline(tstat_hot_1,color="r")
        plt.axvline(tstat_hot_2,color="r")
        self.slabevaldicts.append(miscdict)
      ofp.close()
    plt.tight_layout()
    plt.savefig(figname,dpi=200)
    if not noshow:
      plt.show()
    return
  
  # creates sequential elements in a specific direction based on their positions
  # also calculates averaged temperatures
  # boundarycloseness is used to specify distance from pbc where the reference cell is used for length determination instead
  def get_1D_elements(self,direc,tol,boundarycloseness = 5,fullref=False):
    considered = [False] * self.natoms
    dirlen = self.geo.coors[:,direc]
    self.oneDlens[direc]       = []
    self.oneDels[direc]        = []
    self.oneDtemps[direc]      = []
    self.oneDlenerrs[direc]    = []
    self.oneDtemperrs[direc]   = []
    deviation_file = False
    df = None
    for i, dist in enumerate(dirlen):
      if not considered[i]:
        el = self.geo.elems[i]
        logical = ((dirlen>(dist-tol)) & (dirlen<(dist+tol))) & (el == self.geo.elems)
        considered |= logical
        atoms    = self.coors[direc][logical]
        atemps   = self.temps[logical]
        aids     = np.asarray(range(len(self.temps)))[logical]
        numats   = len(atemps)
        if numats>1:
          statlen = (numats-1)**0.5
        else:
          statlen = 1
        if numats != 0:
            self.oneDels[direc].append(el)
            if ((atoms[0] < boundarycloseness) | (atoms[0] > (self.geo.cell[direc][direc]-boundarycloseness))) | fullref:
              atoms = self.geo.coors[logical,direc]
            self.oneDlens[direc].append(np.mean(atoms))
            self.oneDlenerrs[direc].append(np.std(atoms)/statlen)
            self.oneDtemps[direc].append(np.mean(atemps))
            self.oneDtemperrs[direc].append(np.std(atemps)/statlen)
            # HACK to detect incorrectly sorted reference files and to give a resort guide
            if self.oneDtemperrs[direc][-1] > 1.0:
              if not deviation_file:
                deviation_file = True
                df = open("l_deviations.dat","w")
              spreadtemp = self.oneDtemps[direc][-1]
              actatoms = self.coors[direc][logical]
              print (np.mean(atemps),el,np.mean(atoms),self.oneDlenerrs[direc][-1],self.oneDtemperrs[direc][-1],sum(logical))
              ltemplog = atemps<spreadtemp
              htemplog = atemps>spreadtemp
              llens = actatoms[ltemplog]
              hlens = actatoms[htemplog]
              lreflens = atoms[ltemplog]
              hreflens = atoms[htemplog]
              lids  = aids[ltemplog]
              hids  = aids[htemplog]
              ltemps = atemps[ltemplog]
              htemps = atemps[htemplog]
              df.write("# ref mean coordinate; element; mean temperature; mean low temperature; mean high temperature; min temperature; max temperature\n")
              df.write("%12.6f %s %12.6f %12.6f %12.6f %12.6f %12.6f\n" % (np.mean(atoms),el,np.mean(atemps),np.mean(atemps[ltemplog]),
                                                                           np.mean(atemps[htemplog]),np.min(atemps[ltemplog]),np.max(atemps[htemplog])))
              df.write("# atom id; sim len; ref len; low temp\n")
              for j in range(len(llens)):
                df.write("%6d %12.6f %12.6f %12.6f\n"%(lids[j],llens[j],lreflens[j],ltemps[j]))
              df.write("# atom id; sim len; ref len; high temp\n")
              for j in range(len(hlens)):
                df.write("%6d %12.6f %12.6f %12.6f\n"%(hids[j],hlens[j],hreflens[j],htemps[j]))
        else:
            self.oneDels[direc].append(el)
            self.oneDlens[direc].append(0)
            self.oneDlenerrs[direc].append(0)
            self.oneDtemps[direc].append(0)
            self.oneDtemperrs[direc].append(0)
          #print np.mean(atemps[ltemplog])
          #print np.mean(atemps[htemplog])
    if deviation_file:
      df.close()
    self.oneDlens[direc]       = np.asarray(self.oneDlens[direc]    )
    self.oneDels[direc]        = np.asarray(self.oneDels[direc]     )
    self.oneDtemps[direc]      = np.asarray(self.oneDtemps[direc]   )
    self.oneDlenerrs[direc]    = np.asarray(self.oneDlenerrs[direc] )
    self.oneDtemperrs[direc]   = np.asarray(self.oneDtemperrs[direc])
    return
  
  # finds the interfaces based on edge atoms in the reference structure
  # one dimensional structure in that direction has to be done first !
  # nlconf: indicates outer elements enclosing the section, the first one is the initial segment
  # nlnames: names for all sections
  # nlcolors: plot colors for sections
  # direc: direction to be investigated
  # minsegatoms: minimum number of atoms in the segments
  # namindist: minimum distance of atoms for segment detection
  # refmode: take coordinates from reference file instead
  #
  # returns: list of all section boundaries and list with section types
  def find_interfaces(self,nlconf,nlnames,nlcolors,direc,minsegatoms=0,namindist=0,refmode=False):
    
    self.nlcolors = nlcolors
    if refmode:
      sortlog = self.geo.coors[direc].argsort()
      selems  = self.geo.elems[sortlog]
      slens   = self.geo.coors[direc][sortlog]
    else:
      sortlog = self.oneDlens[direc].argsort()
      selems  = self.oneDels[direc][sortlog]
      slens   = self.oneDlens[direc][sortlog]
    self.seglist    = []
    self.segnames   = []
    self.segcolors  = []
    self.segside    = []
    
    maxlen = self.geo.cell[direc][direc]
    midlen = maxlen/2
    
    currpos = 0
    natcurrseg = 0
    lmidlen = 0
    for eid,el in enumerate(selems):
      if (eid + 1) != len(selems):
        if (currpos + 1) == len(nlconf):
          nextpos = 0
        else:
          nextpos = currpos + 1
        if len(nlconf[currpos]) == 2:
          if ((nlconf[currpos][1] == el) & 
                  (selems[eid+1] in nlconf[nextpos][0].split("|")) &
                  (natcurrseg > minsegatoms) &
                  ((slens[eid+1]-slens[eid]) > namindist)):
            natcurrseg = 0
            
            rmidlen = (slens[eid+1]-slens[eid])/2.0 + slens[eid]
            
            # check on which side of plot:
            if lmidlen < midlen:
              self.segside.append(1)
            elif lmidlen >= midlen:
              self.segside.append(-1)
            
            self.segnames.append(nlnames[currpos])
            self.segcolors.append(nlcolors[currpos])
            currpos = nextpos
            self.seglist.append([lmidlen,rmidlen])
            lmidlen = rmidlen
        elif len(nlconf[currpos]) == 1:
          
          if ((el in nlconf[currpos][0].split("|")) &
              (selems[eid+1] in nlconf[nextpos][0].split("|"))
              ):
                  #& ((slens[eid+1]-slens[eid]) > namindist)):
            natcurrseg = 0
            
            rmidlen = (slens[eid+1]-slens[eid])/2.0 + slens[eid]
            
            # check on which side of plot:
            if lmidlen < midlen:
              self.segside.append(1)
            elif lmidlen >= midlen:
              self.segside.append(-1)
            
            self.segnames.append(nlnames[currpos])
            self.segcolors.append(nlcolors[currpos])
            currpos = nextpos
            self.seglist.append([lmidlen,rmidlen])
            lmidlen = rmidlen
            
        natcurrseg += 1
      
    self.seglist    = np.asarray(self.seglist          )
    self.segnames   = np.asarray(self.segnames          )
    self.segcolors  = np.asarray(self.segcolors          )
    self.segside    = np.asarray(self.segside          )
    return self.seglist
  
  # fits the interfaces previously found
  def fit_interfaces(self,direc,icross="",icrossignoreels = [],ignoreels = []):
    self.fitparams = []
    self.icrosslog = [False] * len(self.oneDels[direc])
    for sid,seg in enumerate(self.seglist):
      llen = seg[0]
      rlen = seg[1]
      hfdirec = self.oneDlens[direc]
      
      if icross == self.segnames[sid]:
        logical = self.get_logical_in_segment(seg,hfdirec,direc,self.oneDels[direc],ignoreels=np.concatenate((icrossignoreels,ignoreels)))
      else:
        logical = self.get_logical_in_segment(seg,hfdirec,direc,self.oneDels[direc],ignoreels=ignoreels)
      self.icrosslog = self.icrosslog | logical
      
      if sum(logical!=0):
        fit = np.polyfit(hfdirec[logical],self.oneDtemps[direc][logical],1)
      else:
        print ("WARNING: no valid atoms from %f to %f" % (llen,rlen))
        fit = [0.0, 0.0]
      fit_fn = np.poly1d(fit)
      self.fitparams.append(fit_fn)
    return self.fitparams
  
  # finds the segments not influenced by thermostat and returns a logical array
  def find_tstat_influence(self,direc,tstat=25):
    seglog = [False] * len(self.seglist)
    
    maxlen = self.geo.cell[direc][direc]
    midlen = maxlen/2
    middle_left = midlen-tstat/2
    middle_right = midlen+tstat/2
    left = tstat/2
    right = maxlen - tstat/2
    
    
    
    for sid, seg in enumerate(self.seglist):
      rmidlen = seg[1]
      lmidlen = seg[0]
      if ((lmidlen > left) & (rmidlen < middle_left)) | ((lmidlen > middle_right) & (rmidlen < right)): 
        seglog[sid] = True
    return np.asarray(seglog)
  
  # returns a logical in a certain segment for a specific array
  def get_logical_in_segment(self,seg,arr,direc,ellist,ignoreels=[]):
    logical = ((arr < seg[1]) & (arr > seg[0]))
    for iel in ignoreels:
      logical &= (ellist != iel)
    return logical
  
  # get the length intervals describing the bulk of the system between the thermostat boundaries
  def get_tstat_intervals(self,direc,tstat):
    tot_len = self.geo.cell[direc,direc]
    ts_half = tstat/2.
    Tstat_regions = [[ts_half,tot_len/2.-ts_half],[tot_len/2.+ts_half,tot_len-ts_half]]
    return np.asarray(Tstat_regions)
  
  # gets the cross sectional area perpendicular to the direction specified
  # only works for a rectangular cell
  def get_cross_area(self,direc):
    if direc == 0:
      area = self.geo.cell[1,1]*self.geo.cell[2,2]
    if direc == 1:
      area = self.geo.cell[0,0]*self.geo.cell[2,2]
    if direc == 2:
      area = self.geo.cell[1,1]*self.geo.cell[0,0]
    #print("cross area",area)
    return area
  
  
  def plot_full_profile(self,direc):
    a0 = 1.43
    self.visualize_elems(self.oneDlens[direc]/a0,self.oneDtemps[direc],self.oneDels[direc],labels=True)
    return
  
  def plot_raw_profile(self,direc):
    a0 = 1.0
    plt.scatter(self.coors[direc]/a0,self.temps)
    return
  
  # plots the temperature profile of the NEMD run by using the averaged atom positions and temperatures
  def plot_element_profile(self,direc, monotonecol=False, ignoreels = [],markersize=8, icirc = False, icirccol = "k", 
                           icross="", icrossignoreels=["C","H"]):
    if icirc:
        mecol = icirccol
        mewd = 3
    else:
        mecol = icirccol
        mewd = 0
    a0 = 1.0
    if icross != "":
      hfdirec = self.oneDlens[direc]
      slogical = [False] * len(self.oneDels[direc])
      for sid,seg in enumerate(self.seglist):
        if icross == self.segnames[sid]:
          slogical |= self.get_logical_in_segment(seg,hfdirec,direc,self.oneDels[direc],ignoreels=np.concatenate((icrossignoreels,ignoreels)))
        else:
          slogical |= self.get_logical_in_segment(seg,hfdirec,direc,self.oneDels[direc],ignoreels=ignoreels)
    
    for el in self.systemels:
      logical = (np.asarray(self.oneDels[direc]) == el)
      if icross != "":
        logical &= slogical
      if (el in self.defaultcols.keys()) & (not monotonecol):
        plcol = self.defaultcols[el]
      else:
        plcol = "k"
      if el not in ignoreels:
        plt.plot((self.oneDlens[direc][logical])/a0,self.oneDtemps[direc][logical],"o",color=plcol,markersize=markersize,markeredgecolor = mecol, markeredgewidth = mewd)
    return
  
  
  def plot_savgol(self, direc, window, polyorder, ignoreels = [], icross="", icrossignoreels=["C","H"]):
    
    hfdirec = self.oneDlens[direc]
    logical = [False] * len(self.oneDels[direc])
    for sid,seg in enumerate(self.seglist):
      if icross == self.segnames[sid]:
        logical |= self.get_logical_in_segment(seg,hfdirec,direc,self.oneDels[direc],ignoreels=np.concatenate((icrossignoreels,ignoreels)))
      else:
        logical |= self.get_logical_in_segment(seg,hfdirec,direc,self.oneDels[direc],ignoreels=ignoreels)
    tempdata = self.oneDtemps[direc][logical]
    xdata = self.oneDlens[direc][logical]
    sarr = np.argsort(xdata)
    xdata = xdata[sarr]
    tempdata = tempdata[sarr]
    svgdata = smooth.non_uniform_savgol(xdata,tempdata,window,polyorder)
    #svgdata = scipy.signal.savgol_filter(tempdata,window,polyorder)
    plt.plot(xdata,svgdata)
    return
  
  # same as plot_savgol but using the raw data instead of the averaged one 
  def plot_savgol_raw(self, direc, window, polyorder, ignoreels = [], icross="", icrossignoreels=["C","H"]):
    
    hfdirec = self.oneDlens[direc]
    logical = [False] * len(self.geo.coors[:,direc])
    for sid,seg in enumerate(self.seglist):
      if icross == self.segnames[sid]:
        logical |= self.get_logical_in_segment(seg,self.coors[direc],direc,self.geo.elems,ignoreels=np.concatenate((icrossignoreels,ignoreels)))
      else:
        logical |= self.get_logical_in_segment(seg,self.coors[direc],direc,self.geo.elems,ignoreels=ignoreels)
    tempdata = self.temps[logical]
    xdata = self.coors[direc][logical]
    sarr = np.argsort(xdata)
    xdata = xdata[sarr]
    tempdata = tempdata[sarr]
    svgdata = smooth.non_uniform_savgol(xdata,tempdata,window,polyorder)
    #svgdata = scipy.signal.savgol_filter(tempdata,window,polyorder)
    plt.plot(xdata,svgdata)
    #plt.plot(xs,svgdata(xs))
    return
  
  def plot_spline(self, direc, ignoreels = [], icross="", icrossignoreels=["C","H"], smooth=1000, lw=3):
    
    hfdirec = self.oneDlens[direc]
    logical = [False] * len(self.geo.coors[:,direc])
    for sid,seg in enumerate(self.seglist):
      if icross == self.segnames[sid]:
        logical |= self.get_logical_in_segment(seg,self.coors[direc],direc,self.geo.elems,ignoreels=np.concatenate((icrossignoreels,ignoreels)))
      else:
        logical |= self.get_logical_in_segment(seg,self.coors[direc],direc,self.geo.elems,ignoreels=ignoreels)
    tempdata = self.temps[logical]
    xdata = self.coors[direc][logical]
    sarr = np.argsort(xdata)
    xdata = xdata[sarr]
    tempdata = tempdata[sarr]
    splinefunc = scipy.interpolate.UnivariateSpline(xdata,tempdata,k=3,s=smooth)
    xs = np.linspace(min(xdata),max(xdata),10000)
    plt.plot(xs,splinefunc(xs),linewidth=lw)
    return
  
  def evaluate_tloc(self,loccoors,loctemps,valnum=1000,tlocrange=1.5,minval=None,maxval=None):
    if minval is None:
      minval = min(loccoors)
    if maxval is None:
      maxval = max(loccoors)
    valrange = np.linspace(minval,maxval,valnum)
    loctemp  = np.zeros(np.shape(valrange))
    for i,val in enumerate(valrange):
      llim = (val - tlocrange)
      hlim = (val + tlocrange)
      loclog = (loccoors > llim) & (loccoors < hlim)
      loctemp[i] = (np.mean(loctemps[loclog]))
      if np.isnan(loctemp[i]):
        if i == 0:
          loctemp[i] = min(loctemps)
        else:
          loctemp[i] = loctemp[i-1]
    return valrange, loctemp
  
  # plots local temperature based on result data
  def plot_tloc_profile(self, direc, ignoreels=[], valnum=1000, tlocrange = 1.5,offset=0,linewidth=3):
    logical = [True]*len(self.geo.coors[:,direc])
    for el in ignoreels:
      logical &= (self.geo.elems != el)
    loccoors = self.geo.coors[:,direc][logical]
    loctemps = self.temps[logical]
    valrange,loctemp = self.evaluate_tloc(loccoors,loctemps,valnum=valnum,tlocrange=tlocrange,minval=0.0,maxval=self.geo.cell[direc,direc])
  
    plt.plot(valrange,loctemp+offset,'k-', linewidth=linewidth)
    return loctemp
  
  def plot_vlines(self):
    for sid,seg in enumerate(self.seglist):
      plt.axvspan(seg[0], seg[1], alpha=0.2, color=self.segcolors[sid])
    return
  
  
  def plot_resistance_profile(self,direc,window,polyorder,smoothfact, nplotvals=10000, ignoreels = [], icross="", 
                              icrossignoreels=["C","H"], useflux=90,n_convtest=100, fontsize=15, mode="RA", rhoN = 4,
                              tstat = 50, elslope = False, printraw = True, smoothmode="SG",tlocrange=1.5, refcoors=False):
    '''
        Several modes:  RA  ... resistance times A
                        R   ... thermal resistance
                        rho ... resistance times N (number of linkers in heat transport direction for one cell)
        elslope ... specifies whether or not the slopes in the averaged element profile are to be evaluated and displayed
        printraw ... print the raw data overlaid for all the segments outside of the thermostat influence
    '''
    
    if elslope:
      sortx_order = np.argsort(self.oneDlens[direc])
      hfdirec = self.oneDlens[direc][sortx_order]
      eltemps = self.oneDtemps[direc][sortx_order]
      elslopexvals = []
      elslopetemps = []
      elslopes = []
      for hid, hel in enumerate(hfdirec):
        if hid != 0:
          halfx = (hfdirec[hid]-hfdirec[hid-1])/2
          halftemp = (eltemps[hid]-eltemps[hid-1])/2
          elslopexvals.append(hfdirec[hid]+halfx)
          elslopetemps.append(eltemps[hid]+halftemp)
          elslopes.append(halftemp/halfx)
      elslopexvals = np.asarray(elslopexvals)
      elslopetemps = np.asarray(elslopetemps)
      elslopes     = np.asarray(elslopes    )
    if refcoors:
      evcoors = self.geo.coors[:,direc]
    else:
      evcoors = self.coors[direc]
    logical = [False] * len(self.geo.coors[:,direc])
    for sid,seg in enumerate(self.seglist):
      if icross == self.segnames[sid]:
        logical |= self.get_logical_in_segment(seg,self.geo.coors[:,direc],direc,self.geo.elems,ignoreels=np.concatenate((icrossignoreels,ignoreels)))
      else:
        logical |= self.get_logical_in_segment(seg,self.geo.coors[:,direc],direc,self.geo.elems,ignoreels=ignoreels)
    tempdata = self.temps[logical]
    xdata = evcoors[logical]
    sarr = np.argsort(xdata)
    xdata = xdata[sarr]
    tempdata = tempdata[sarr]
    
    if smoothmode == "tloc":
      xtl,ytl = self.evaluate_tloc(xdata,tempdata,valnum=nplotvals,tlocrange=tlocrange,minval=0.0,maxval=self.geo.cell[direc,direc])
      
      outfilename = "avg_res_profile_tloc_"+str(tlocrange)+".dat"
      #xs = xtl
      #ys = ytl
      
      # further smoothen with spline
      splinefunc = scipy.interpolate.UnivariateSpline(xtl,ytl,s=smoothfact,k=3)
      xs = np.linspace(0.,self.geo.cell[direc,direc],nplotvals)
      ys = splinefunc(xs)
      
      # further smoothen with GPR
      #ylowoffset = min(ytl)
      #scaley = max(ytl)-min(ytl)
      #kernel = WhiteKernel() + RBF() * C()
      #xdata_2d = np.atleast_2d(xtl).T
      #gpr = GaussianProcessRegressor(kernel=kernel, random_state=0)
      #gpr.fit(xdata_2d, (ytl-ylowoffset)/scaley)
      #gpr_params = gpr.kernel_.get_params()
      #xarr = np.atleast_2d(np.linspace(0.,self.geo.cell[direc,direc], nplotvals)).T
      #yarr, sigma = gpr.predict(xarr, return_std=True)
      #xs = xarr[:,0]
      #ys = yarr*scaley+ylowoffset
      
    elif smoothmode == "SG":
      svgdata = smooth.non_uniform_savgol(xdata,tempdata,window,polyorder)
      splinefunc = scipy.interpolate.UnivariateSpline(xdata,svgdata,s=smoothfact,k=3)
      xs = np.linspace(0.,self.geo.cell[direc,direc],nplotvals)
      ys = splinefunc(xs)
      outfilename = "avg_res_profile_sg_"+str(window)+"_"+str(polyorder)+"_spline"+str(smoothfact)+".dat"
    elif smoothmode == "spline":
      splinefunc = scipy.interpolate.UnivariateSpline(xdata,tempdata,s=smoothfact,k=3)
      xs = np.linspace(0.,self.geo.cell[direc,direc],nplotvals)
      ys = splinefunc(xs)
      outfilename = "avg_res_profile_spline"+str(smoothfact)+".dat"
      #residual = splinefunc.get_residual()
      #knotpos = splinefunc.get_knots()  
    elif smoothmode == "GPR":
      #gp = GaussianProcessRegressor(kernel=kernel, n_restarts_optimizer=9)

      ## Fit to data using Maximum Likelihood Estimation of the parameters
      #gp.fit(X, y)

      ## Make the prediction on the meshed x-axis (ask for MSE as well)
      #y_pred, sigma = gp.predict(x, return_std=True)

      #kernel = DotProduct() + WhiteKernel()
      ylowoffset = min(tempdata)
      scaley = max(tempdata)-min(tempdata)
      kernel = WhiteKernel() + RBF() * C()
      xdata_2d = np.atleast_2d(xdata).T
      gpr = GaussianProcessRegressor(kernel=kernel, random_state=0)
      gpr.fit(xdata_2d, (tempdata-ylowoffset)/scaley)
      gpr_params = gpr.kernel_.get_params()
      xarr = np.atleast_2d(np.linspace(0.,self.geo.cell[direc,direc], nplotvals)).T
      yarr, sigma = gpr.predict(xarr, return_std=True)
      xs = xarr[:,0]
      ys = yarr*scaley+ylowoffset
      outfilename = "avg_res_profile_gpr.dat"
      
      # write the parameters in a file
      gfp = open("gpr.param","w")
      gfp.write("Gaussian Process Regression Parameters\n")
      gfp.write("Kernel: WhiteKernel() + RBF() * C()\n")
      gfp.write("params:\n")
      gfp.write(str(gpr_params))
      gfp.close()
      
    
    
    # here calculate the error from the filtering
    #svgerr = util.rmse(svgdata,tempdata)
    #print(svgerr)
    
    #svgdata = scipy.signal.savgol_filter(tempdata,window,polyorder)
    #plt.plot(xdata,svgdata)
    
    
    
    # debug lines to plot the knots in the spline
    #plt.vlines(knotpos,250,350)
    #print(len(xdata),residual)
    
    fluxdists = np.asarray(list(map(int,np.floor(np.linspace(0,len(self.fluxt)-1,n_convtest)))))
    favg = np.zeros(np.shape(fluxdists))
    for i,fdist in enumerate(fluxdists):
      favg[i] = np.mean(self.fluxt[fdist:])
    
    
    tgrad = np.gradient(ys,xs)
    uphalf = xs>(self.geo.cell[direc,direc]/2)
    tgrad[uphalf] = -tgrad[uphalf]
    
    if mode == "RA":
      resprof = tgrad/favg[useflux]/1e-9
    elif mode == "R":
      resprof = tgrad/favg[useflux]/self.get_cross_area(direc)*1e20/1e-9
    elif mode == "rho":
      resprof = tgrad/favg[useflux]/self.get_cross_area(direc)*1e20*rhoN/1e-9
    
    intervals = self.get_tstat_intervals(direc,tstat)
    leftbulk = (xs < intervals[0,1]) & (xs > intervals[0,0])
    rightbulk = (xs < intervals[1,1]) & (xs > intervals[1,0])
    totalbulk = leftbulk | rightbulk
    bulkmax = max(resprof[totalbulk])
    plt.scatter(xs,ys,c=resprof,cmap="inferno",s=50,vmin=0,vmax=bulkmax)
    cb = plt.colorbar()
    
    if mode == "RA":
      cb.set_label("RA / m$\mathbf{^2}$K/GW", fontsize=fontsize+3, fontweight="bold")
    elif mode == "R":
      cb.set_label("R / K/GW", fontsize=fontsize+3, fontweight="bold")
    elif mode == "rho":
      cb.set_label("$\mathbf{ {\\rho}\: /\: K/GW}$", fontsize=fontsize+3, fontweight="bold")
    cb.ax.tick_params(labelsize=fontsize)
    
    # segment to plot the segment overlaid temperature profile
    figmain = plt.gcf()
    
    fig2 = plt.figure(11,figsize=(12,8))
    
    seglog = self.find_tstat_influence(direc,tstat)
    cb2 = None
    firstseg = None
    lastseg = None
    fullseglistx = []
    fullseglisty = []
    fullseglistresprof = []
    amountsegs = []
    segmentstartendvalues = []
    found_first_in_half = False
    halflen = self.geo.cell[direc,direc]/2
    if printraw:
      rawfp = open("raw_temp_prof_repeat_unit.dat","w")
      rawx = []
      rawtemp = []
      rawels = []
    print("segment lengths:")
    for sid, seg in enumerate(self.seglist):
      if seglog[sid]:
        print(seg[1]-seg[0])
        if printraw:
          if icross == self.segnames[sid]:
            rawlog = self.get_logical_in_segment(seg,self.coors[direc],direc,self.geo.elems,ignoreels=np.concatenate((icrossignoreels,ignoreels)))
          else:
            rawlog = self.get_logical_in_segment(seg,self.coors[direc],direc,self.geo.elems,ignoreels=ignoreels)
          rawxcurr = self.coors[direc][rawlog]
          rawtempcurr = self.temps[rawlog]
          rawelscurr = self.geo.elems[rawlog]
        if (seg[0] > halflen) & (firstseg != lastseg):
          firstseg = lastseg
          found_first_in_half = False
        xv_in_seg_log = (xs < seg[1]) & (xs > seg[0])
        segprofx = xs[xv_in_seg_log]
        segprofy = ys[xv_in_seg_log]
        if firstseg is None:
          firstseg = self.segnames[sid]
        if self.segnames[sid] == firstseg:
          if lastseg is None:
            lastseg = self.segnames[sid-1]
          prevsegx = 0
          prevsegy = 0
          minxcurr = min(xs[xv_in_seg_log])
          minycurr = min(ys[xv_in_seg_log])
          segprofx -=  minxcurr
          segprofy -=  minycurr
          minswap = 0.
          if (seg[0] < intervals[1,1]) & (seg[0] > intervals[1,0]):
            len_next_seg = self.seglist[sid+1][1]-self.seglist[sid+1][0]
            minswap = min(-segprofx)-len_next_seg
            segprofx = -segprofx-minswap
          fullseglistx.append(segprofx)
          fullseglisty.append(segprofy)
          fullseglistresprof.append(resprof[xv_in_seg_log])
          amountsegs.append(1)
          segmentstartendvalues.append([seg[0],seg[1]])
          found_first_in_half = True
          if printraw & found_first_in_half:
            rawx.append(rawxcurr)
            rawtemp.append(rawtempcurr)
            rawels.append(rawelscurr)
            rawx[-1] -= minxcurr
            rawtemp[-1] -= minycurr
            if (seg[0] < intervals[1,1]) & (seg[0] > intervals[1,0]):
              rawx[-1] = -rawx[-1]-minswap
        elif found_first_in_half:
          segprofx -= minxcurr
          segprofy -= minycurr
          if (seg[0] < intervals[1,1]) & (seg[0] > intervals[1,0]):
            segprofx = -segprofx-minswap
            minsecondseg = min(ys[xv_in_seg_log])
          fullseglistx[-1] = np.concatenate((fullseglistx[-1],segprofx))
          fullseglisty[-1] = np.concatenate((fullseglisty[-1],segprofy))
          fullseglistresprof[-1] = np.concatenate((fullseglistresprof[-1],resprof[xv_in_seg_log]))
          segmentstartendvalues[-1][1] = seg[1]
          amountsegs[-1] += 1
          if printraw & found_first_in_half:
            rawxcurr -= minxcurr
            rawtempcurr -= minycurr
            if (seg[0] < intervals[1,1]) & (seg[0] > intervals[1,0]):
              rawxcurr = -rawxcurr-minswap
            rawx[-1]   = np.concatenate((rawx[-1]   ,rawxcurr)   )
            rawtemp[-1]= np.concatenate((rawtemp[-1],rawtempcurr))
            rawels[-1] = np.concatenate((rawels[-1] ,rawelscurr) )
          
            
          
        
    standard_amountsegs = amountsegs[0]
    halfxnum = int(len(fullseglistx[0])/2)
    baseyoffset = fullseglisty[0][halfxnum]
    if printraw:
      halfrawxnum = int(len(rawx[0])/2)
      baserawyoffset = rawx[0][halfrawxnum]
    for sid, seg in enumerate(fullseglistx):
      if amountsegs[sid] == standard_amountsegs:
        yoffset = fullseglisty[sid][halfxnum]-baseyoffset
        fullseglisty[sid] -= yoffset
        plt.scatter(fullseglistx[sid],fullseglisty[sid],c=fullseglistresprof[sid],cmap="inferno",s=50,vmin=0,vmax=bulkmax)
        if printraw:
          rawyoffset = rawtemp[sid][halfrawxnum]-baserawyoffset
          rawtemp[sid] -= rawyoffset
          for rid,rawval in enumerate(rawx[sid]):
            rawfp.write("%12.6f %3s %12.6f\n" % (rawval, rawels[sid][rid], rawtemp[sid][rid]))
        # debugging text to identify the segment positions
        #plt.text(fullseglistx[sid][100],fullseglisty[sid][100]-yoffset,str(segmentstartvalues[sid]),fontsize=fontsize)
        #print(np.median(fullseglisty[sid]))
        if cb2 is None:
          cb2 = plt.colorbar()
          
          
    if printraw:
      rawfp.close()
        
          
    if cb2 is not None:
      if mode == "RA":
        cb2.set_label("RA / m$\mathbf{^2}$K/GW", fontsize=fontsize+3, fontweight="bold")
      elif mode == "R":
        cb2.set_label("R / K/GW", fontsize=fontsize+3, fontweight="bold")
      elif mode == "rho":
        cb2.set_label("$\mathbf{ {\\rho}\: /\: K/GW}$", fontsize=fontsize+3, fontweight="bold")
    plt.xlabel("$\mathbf{z}$ [$\mathbf{\AA}$]", fontweight="bold", fontsize=fontsize)
    plt.ylabel("$\mathbf{\overline{T} ^\prime}$ [K]", fontweight="bold", fontsize=fontsize)  
    plt.tight_layout()
    plt.savefig("overlay_res_temp_prof.png")
    
    # remove all entries from the list with an incorrect amountsegs
    toremovefromlist = []
    for sid, seg in enumerate(fullseglistx):
      if amountsegs[sid] != standard_amountsegs:
        toremovefromlist.append(sid)
      else:
        sarr = np.argsort(seg)
        fullseglistx[sid] = fullseglistx[sid][sarr]
        fullseglisty[sid] = fullseglisty[sid][sarr]
        fullseglistresprof[sid] = fullseglistresprof[sid][sarr]
    toremovefromlist.sort(reverse=True)
    for index in toremovefromlist:
      fullseglistx.pop(index)
      fullseglisty.pop(index)
      fullseglistresprof.pop(index)
    
    fullseglistx       = np.asarray(fullseglistx      )
    fullseglisty       = np.asarray(fullseglisty      )
    fullseglistresprof = np.asarray(fullseglistresprof)
    
    plt.figure(12,figsize=(12,16))
    ax1 = plt.subplot(211)
    
    # average the elslope data for the segments
    if elslope:
      elsx = []
      elst = []
      elss = []
      #print (elslopexvals)
      for sid,seg in enumerate(segmentstartendvalues):
        if amountsegs[sid] == standard_amountsegs:
          logical = (elslopexvals < seg[1]) & (elslopexvals > seg[0])
          if (seg[0] < intervals[1,1]) & (seg[0] > intervals[1,0]):
            elsx.append(-elslopexvals[logical]+min(elslopexvals[logical])+(seg[1]-seg[0]))
          else:
            elsx.append(elslopexvals[logical]-min(elslopexvals[logical]))
          elst.append(elslopetemps[logical])
          elss.append(elslopes[logical]    )
          sarr = np.argsort(elsx[-1])
          elsx[-1] = elsx[-1][sarr]
          elst[-1] = elst[-1][sarr]
          elss[-1] = elss[-1][sarr]
      elsx = np.asarray(elsx)
      elst = np.asarray(elst)
      elss = np.asarray(elss)
      #print(elsx)
      elsxmean = np.mean(elsx,axis=0)
      elstmean = np.mean(elst,axis=0)
      elssmean = np.mean(elss,axis=0)
      if mode == "RA":
        elsresmean = elssmean/favg[useflux]/1e-9
      elif mode == "R":
        elsresmean = elssmean/favg[useflux]/self.get_cross_area(direc)*1e20/1e-9
      elif mode == "rho":
        elsresmean = elssmean/favg[useflux]/self.get_cross_area(direc)*1e20*rhoN/1e-9
        
    
    # average the actual slopes
    data_lens = []
    for eid,entry in enumerate(fullseglisty):
      data_lens.append(len(entry))
    # cut away excess values that occured due to segmentation
    lowlen = min(data_lens)
    for eid,entry in enumerate(fullseglisty):
      fullseglisty[eid] = list(fullseglisty[eid][:lowlen])
      fullseglistresprof[eid] = list(fullseglistresprof[eid][:lowlen])
      fullseglistx[eid] = list(fullseglistx[eid][:lowlen])
    fullseglisty        = np.asarray(list(fullseglisty)      )
    fullseglistresprof  = np.asarray(list(fullseglistresprof))
    fullseglistx        = np.asarray(list(fullseglistx)      )
    print(np.shape(fullseglisty))
    meansegy = np.mean(fullseglisty,axis=0)
    meansegyerr = np.std(fullseglisty,axis=0)/np.sqrt(len(fullseglisty[:,0]))
    meansegres = np.mean(fullseglistresprof,axis=0)
    meansegreserr = np.std(fullseglistresprof,axis=0)/np.sqrt(len(fullseglistresprof[:,0]))
    
    plt.scatter(fullseglistx[0],meansegy,c=meansegres,cmap="inferno",s=50,vmin=0,vmax=max(meansegres))
    #plt.scatter(fullseglistx[0],meansegy+meansegyerr,c=meansegres,cmap="inferno",s=10,vmin=0,vmax=max(meansegres),alpha=0.2)
    #plt.scatter(fullseglistx[0],meansegy-meansegyerr,c=meansegres,cmap="inferno",s=10,vmin=0,vmax=max(meansegres),alpha=0.2)
    cb3 = plt.colorbar()
    if mode == "RA":
      cb3.set_label("RA / m$\mathbf{^2}$K/GW", fontsize=fontsize+3, fontweight="bold")
    elif mode == "R":
      cb3.set_label("R / K/GW", fontsize=fontsize+3, fontweight="bold")
    elif mode == "rho":
      cb3.set_label("$\mathbf{ {\\rho}\: /\: K/GW}$", fontsize=fontsize+3, fontweight="bold")
    ax1.fill_between(fullseglistx[0], meansegy-meansegyerr, meansegy+meansegyerr,zorder=-1,alpha=0.2)
    plt.xlabel("$\mathbf{z}$ [$\mathbf{\AA}$]", fontweight="bold", fontsize=fontsize)
    plt.ylabel("$\mathbf{\overline{T} ^\prime}$ [K]", fontweight="bold", fontsize=fontsize)
    ax2 = plt.subplot(212)
    
    plt.scatter(fullseglistx[0],meansegres,c=meansegy,cmap="plasma",s=50,vmin=min(meansegy),vmax=max(meansegy))
    #plt.scatter(fullseglistx[0],meansegres+meansegreserr,c=meansegy,cmap="plasma",s=10,vmin=min(meansegy),vmax=max(meansegy),alpha=0.2)
    #plt.scatter(fullseglistx[0],meansegres-meansegreserr,c=meansegy,cmap="plasma",s=10,vmin=min(meansegy),vmax=max(meansegy),alpha=0.2)
    cb4 = plt.colorbar()
    cb4.set_label("$\mathbf{\overline{T} ^\prime}$ [K]", fontsize=fontsize+3, fontweight="bold")
    ax2.fill_between(fullseglistx[0], meansegres-meansegreserr, meansegres+meansegreserr,zorder=-1,alpha=0.2)
    if mode == "RA":
      plt.ylabel("RA / m$\mathbf{^2}$K/GW", fontweight="bold", fontsize=fontsize)
    elif mode == "R":
      plt.ylabel("R / K/GW", fontweight="bold", fontsize=fontsize)
    elif mode == "rho":
      plt.ylabel("$\mathbf{ {\\rho}\: /\: K/GW}$", fontweight="bold", fontsize=fontsize)
    if elslope:
      plt.scatter(elsxmean,elsresmean,s=75)
    plt.xlabel("$\mathbf{z}$ [$\mathbf{\AA}$]", fontweight="bold", fontsize=fontsize)
    
    
    plt.tight_layout()
    plt.savefig("averaged_res_profile_overlay.png")
    
    # write the averaged resistance profile in a file
    ofp = open(outfilename,"w")
    ofp.write("# %12s %12s %12s %12s %12s\n" % ("xpos","temp","tempstd","res","resstd"))
    for lid,line in enumerate(fullseglistx[0]):
      ofp.write("  %12.6f %12.6f %12.6f %12.6f %12.6f\n" % (line,meansegy[lid],meansegyerr[lid],meansegres[lid],meansegreserr[lid]))
    ofp.close()
    
    
    plt.figure(figmain.number)
    
    
    
    return
  
  
  # plots the step interactions over each other
  # direc             ... direction of heat transport
  # segmentname       ... central segment
  # icross            ... ignore cross linker (segmentname)
  # icrossignoreels   ... these elements build the cross linkers
  # figname           ... output name of figure
  # fontsize          ... base fontsize used in plots
  # allvalfit         ... fit all individual vals instead of averaged ones
  # elementvis        ... visualize elements
  # fullfit           ... fit the entire model at once instead of performing linear fits TODO
  # offset            ... plot offset of linear fits
  # oneside           ... plot both sides of thermostat at once
  # neutralels        ... which elements will be plotted in a neutral color
  # appendices        ... how many interfaces are included
  # visualize_int     ... visualize interface between the individual linear fits
  # xoffset           ... offset of x coordinates for plot
  # yoffset           ... offset of y coordinates for plot
  # vissegavgs        ... visualize average over all segments instead of total spread
  # reseval           ... evaluate thermal resistances based on fits
  # ignoreels         ... always ignore these elements in segments not influenced by icrossignoreels
  # nocentralfit      ... only fits the linker in a hardcoded fashion, the area inbetween is considered as node and interface combined
  # fluxnorm          ... normalize the values by heat flux (for comparison)
  # tempnorm          ... normalize the temperature profiles to be normalized to a set height
  
  def plot_all_ints(self,direc,segmentname,tstat=25,icross="",icrossignoreels=[],
                    figname="all_ints.png",fontsize=15,allvalfit=True,elementvis=True,
                    fullfit=0,offset=1,oneside=True,neutralels=[],appendices=1,
                    visualize_int=True,xoffset=0,yoffset=0,vissegavgs=False, reseval=True,
                    ignoreels=[],nocentralfit=False, fluxnorm=False, tempnorm=1.0, nofitplot=False,
                    nolabel=False):
    
    seglog = self.find_tstat_influence(direc,tstat)
    #print (self.segnames,seglog,segmentname)
    seglog &= (self.segnames == segmentname)
    a0 = 1.0
    num_segs = appendices*2+1
    if self.usedflux is None:
      self.set_flux()
      
    directemps = self.oneDtemps[direc]/tempnorm
    if fluxnorm:
      directemps /= self.usedflux
      fluxdiv = self.usedflux*tempnorm
    else:
      fluxdiv = tempnorm
    
    allvals  = []
    allvalst = []
    alllens  = []
    if vissegavgs:
      forvisels = []
      forviscol = []
      forvisx   = []
      forvist   = []
    for appid in range(num_segs):
      alllens.append([[],[]])
      allvals.append([[],[]])
      allvalst.append([[],[]])
      if vissegavgs:
        forvisels.append([[],[]])
        forviscol.append([[],[]])
        forvisx.append([[],[]])
        forvist.append([[],[]])
    allcol   = [""]*num_segs
    
    if allvalfit:
      fitstyle = "-"
    else:
      fitstyle = "--"
    sign = 1
    #print(self.seglist,seglog)
    for sid, seg in enumerate(self.seglist):
      if seglog[sid]:
        sideind = 0
        if self.segside[sid] == 1:
          if not oneside:
            plt.subplot(211)
            sideind = 1
          else:
            sign = 1
        else:
          if not oneside:
            plt.subplot(212)
            sideind = 0
          else:
            sign = -1
        
        # TODO: plot box around
        
        # get means for central segment
        if self.segnames[sid] == icross:
          ignore = icrossignoreels
        else:
          ignore = ignoreels
        logical = self.get_logical_in_segment(seg,self.oneDlens[direc],direc,self.oneDels[direc],ignoreels=ignore)
        meantemp = np.mean(directemps[logical])
        xmean    = np.mean(self.oneDlens[direc][logical])
        
        # loop over all segments
        #print (num_segs)
        for appid in range(num_segs):
          appendice = appid - appendices
          
          if self.segnames[sid+appendice] == icross:
            ignore = icrossignoreels
          else:
            ignore = ignoreels
            
          
          logical = self.get_logical_in_segment(self.seglist[sid+appendice],self.oneDlens[direc],direc,self.oneDels[direc],ignoreels=ignore)
          if elementvis & (not vissegavgs):
            self.visualize_elems(self.oneDlens[direc][logical]-xmean,sign*(directemps[logical]-meantemp),
                                self.oneDels[direc][logical],othercol=self.segcolors[sid+appendice],labels=False,noelcol=neutralels,xoffset=xoffset,yoffset=yoffset)
          elif elementvis & vissegavgs:
            sortx = np.argsort(self.oneDlens[direc][logical])
            forvisels[appid][sideind]  = self.oneDels[direc][logical][sortx]
            forviscol[appid][sideind]  = self.segcolors[sid+appendice]
            forvisx[appid][sideind].append(self.oneDlens[direc][logical][sortx]-xmean  )
            forvist[appid][sideind].append(sign*(directemps[logical][sortx]-meantemp))
            #if len(forvisx[appid][sideind]) == 0:
            #else:
              #forvisx[appid][sideind] += self.oneDlens[direc][logical]-xmean  
              #forvist[appid][sideind] += sign*(directemps[logical]-meantemp)
            self.visualize_elems(self.oneDlens[direc][logical]-xmean,sign*(directemps[logical]-meantemp),
                                self.oneDels[direc][logical],othercol=self.segcolors[sid+appendice],labels=False,
                                noelcol=neutralels,xoffset=xoffset,yoffset=yoffset, alpha=0.05)
          if allvalfit:
            logical = self.get_logical_in_segment(self.seglist[sid+appendice],self.geo.coors[:,direc],direc,self.geo.elems,ignoreels=ignore)
            ctemps = sign*(self.temps[logical]/fluxdiv-meantemp)
            ccoors = self.coors[direc][logical]-xmean
            #ccoors = self.geo.coors[:,direc][logical]-xmean
          else:
            ccoors = self.oneDlens[direc][logical]-xmean
            ctemps = sign*(directemps[logical]-meantemp)
          #print (appid,sideind,len(ccoors))
          allvals[appid][sideind]  = np.concatenate((allvals[appid][sideind],ccoors))
          allvalst[appid][sideind] = np.concatenate((allvalst[appid][sideind],ctemps))
          allcol[appid] = self.segcolors[sid+appendice]
          alllens[appid][sideind].append(self.seglist[sid+appendice][1]-self.seglist[sid+appendice][0])
        
    
    
      
    
    for i in range((not oneside)+1):
      # distance of the node total
      if (len(allvals[appendices+1][i])>0):
        middist = ((min(allvals[appendices+1][i])-max(allvals[appendices][i]))/2+max(allvals[appendices][i])) * 2
      else:
        middist = 0
      if elementvis & vissegavgs:
        for appid in range(num_segs):
          #for el in forvisx[appid][i]:
            #print(len(el))
          xplot = np.mean(np.asarray(forvisx[appid][i]),axis=0)
          #xplot = forvisx[appid][i][0]
          yplot = np.mean(np.asarray(forvist[appid][i]),axis=0)
          self.visualize_elems(xplot,yplot,
                                forvisels[appid][i],othercol=forviscol[appid][i],labels=False,noelcol=neutralels,xoffset=xoffset,yoffset=yoffset)
      if (fullfit == 0) | (fullfit == 5):
        allfits = [[]] * (appendices+1)
        fiterrs = [[]] * (appendices+1)
        fitseglens = [[]] * (appendices+1)
        self.Tdiffs = []
        self.Tdifferrs = []
        for appid in range(num_segs):
          if not ((appid == int(num_segs/2)) & ((nocentralfit)|(fullfit == 5))):
            xrng = np.linspace(min(allvals[appid][i])-offset,max(allvals[appid][i])+offset,300)
            #fit, V = np.polyfit(allvals[appid][i],allvalst[appid][i],1,cov=True)
            # TODO: transfer the linear fit sector including confidence analysis to an external function
            k,d,rvalue,pvalue,stderr = stats.linregress(allvals[appid][i],allvalst[appid][i])
            fit = [k,d]
            
            # stderr from scipy is the standard error of the slope, now calculate stderr of the intercept
            stderr_intercept = stderr*(np.sum(np.asarray(allvals[appid][i])**2)/len(allvals[appid][i]))**0.5
            
            # another problem: for visualization of the slope error have to assume fit through zero -> calculate d offset for a certain k error
            # this should be similar to the error calculated above - use this error instead for fit visualization - it is less biased (assuming the middle is centered)
            # additional consideration: give in a confidence interval by using student t, n-2 degrees of freedom
            student_t = stats.t(len(allvals[appid][i])-2)
            confidence = student_t.ppf(0.99)*stderr
            
            fit_fn = np.poly1d(fit)
            fit_min = np.poly1d([k-confidence,d])
            fit_max = np.poly1d([k+confidence,d])
            
            midx = np.mean(allvals[appid][i])
            dsep_min = fit_min(midx)-fit_fn(midx)
            dsep_max = fit_max(midx)-fit_fn(midx)
            fit_min = np.poly1d([k-confidence,d-dsep_min])
            fit_max = np.poly1d([k+confidence,d-dsep_max])
            
            if (fullfit == 5):
              if min(allvals[appid][i]) < 0:
                  xrng = np.linspace(min(allvals[appid][i])-offset,0,300)
              else:
                  xrng = np.linspace(0,max(allvals[appid][i])+offset,300)
            if not nofitplot:
              plt.plot((xrng+xoffset)/a0,fit_fn(xrng)+yoffset, fitstyle, zorder=5, label="", linewidth=4, color = allcol[appid])
              plt.plot((xrng+xoffset)/a0,fit_min(xrng)+yoffset, fitstyle, zorder=5, label="", linewidth=2, color = allcol[appid])
              plt.plot((xrng+xoffset)/a0,fit_max(xrng)+yoffset, fitstyle, zorder=5, label="", linewidth=2, color = allcol[appid])
              plt.fill_between((xrng+xoffset)/a0,fit_fn(xrng)+yoffset,fit_min(xrng)+yoffset,color = allcol[appid],alpha=0.5)
              plt.fill_between((xrng+xoffset)/a0,fit_fn(xrng)+yoffset,fit_max(xrng)+yoffset,color = allcol[appid],alpha=0.5)
            print (fit_fn, confidence, stderr, stderr_intercept)
            if visualize_int & (appid!=0):
              min_x_val = min(allvals[appid][i])
              linepos = abs(min_x_val-last_xval)/2+min([min_x_val,last_xval])
              self.Tdiffs.append(abs(last_fit(linepos)-fit_fn(linepos)))
              worstcase = (abs(last_fit_max(linepos)-last_fit_min(linepos)) + abs(fit_max(linepos)-fit_min(linepos)))/2
              self.Tdifferrs.append(abs(worstcase))
              if not nofitplot:
                plt.vlines((linepos+xoffset)/a0,fit_fn(linepos)+yoffset,last_fit(linepos)+yoffset,color="C3",linewidth=4,linestyles=(0, (5, 1)))
              
            last_fit = fit_fn
            last_fit_max = fit_max
            last_fit_min = fit_max
            last_xval = max(allvals[appid][i])
            
            # neighboring segments are symmetric - this serves to average over them
            if appid <= num_segs/2:
              allfits[appid] = np.asarray(fit)
              fitseglens[appid] = np.mean(alllens[appid][i])
              fiterrs[appid] = confidence
            else:
              allfits[(num_segs-appid-1)] += np.asarray(fit)
              allfits[(num_segs-appid-1)] /= 2
              fitseglens[(num_segs-appid-1)] += np.mean(alllens[appid][i])
              fitseglens[(num_segs-appid-1)] /= 2
              fiterrs[(num_segs-appid-1)] += confidence
              fiterrs[(num_segs-appid-1)] /= 2
          elif (fullfit==5):
            centind = int(num_segs/2)
            self.Tdiffs.append(0.)
            self.Tdifferrs.append(0.)
            allfits[appid] = np.zeros(2)
            fitseglens[appid] = 0.
            fiterrs[appid] = 0.
              #print allfits,fitseglens
        if nocentralfit:
          centind = int(num_segs/2)
        self.allfits = allfits
        self.fiterrs = fiterrs
        self.fitseglens = fitseglens
        self.allfitcol = allcol
        
        if reseval:
          self.obtain_thermal_resistances(segmentname)
        
      elif fullfit==1:
        fitfunc = interfacefunc
        guesses = [middist,1.0,1.0,1.0]
      elif fullfit==2:
        fitfunc = nointerfacefunc
        guesses = [middist,1.0,1.0]
      elif fullfit==3:
        fitfunc = interfacefunc_fixed_len(middist)
        
        guesses = [1.0,1.0,1.0]
      elif fullfit==4:
        fitfunc = nointerfacefunc_fixed_len(middist)
        guesses = [1.0,1.0]
      if (fullfit != 0)&(fullfit != 5):
        xvals = np.concatenate(np.asarray(allvals)[:,i])
        yvals = np.concatenate(np.asarray(allvalst)[:,i])
        params,cov = optimize.curve_fit(fitfunc,xvals,yvals,p0=guesses)
        
        
        print (params)
        print (cov,np.trace(cov))
        if fullfit == 2:
          self.allfits = [[params[2],0.0],
                          [params[1],params[0]],
                          [params[2],0.0]]
          self.fiterrs = [np.trace(cov)]*3
          totlen = 0
          for entry in np.asarray(alllens)[:,i]:
            totlen += np.mean(entry)
          sideseglen = (totlen-params[0])/2.0
          self.fitseglens = [sideseglen,params[0],sideseglen]
          if reseval:
            self.obtain_ln_thermal_resistances(segmentname)
          self.allfitcol = allcol
          totxdistdiff = abs((max(xvals)-min(xvals))-totlen)/2
          xrng = np.linspace(min(xvals)-totxdistdiff,min(xvals)-totxdistdiff+self.fitseglens[0],1000)
          plt.plot((xrng+xoffset)/a0,fitfunc(xrng,*params)+yoffset,linewidth=4,color=allcol[0])
          xrng = np.linspace(max(xvals)+totxdistdiff-self.fitseglens[0],max(xvals)+totxdistdiff,1000)
          plt.plot((xrng+xoffset)/a0,fitfunc(xrng,*params)+yoffset,linewidth=4,color=allcol[0])
          xrng = np.linspace(min(xvals)-totxdistdiff+self.fitseglens[0],min(xvals)-totxdistdiff+self.fitseglens[0]+self.fitseglens[1],1000)
          plt.plot((xrng+xoffset)/a0,fitfunc(xrng,*params)+yoffset,linewidth=4,color=allcol[1])
        else:
          xrng = np.linspace(min(xvals),max(xvals),1000)
          plt.plot(xrng+xoffset,fitfunc(xrng,*params)+yoffset,linewidth=4)
        
      
          
      
              
      pid = i + 1
      if not oneside:
        plt.subplot(210+pid)
      if not nolabel:
        if fluxnorm:
          plt.ylabel("$\mathbf{\Delta \overline{T}'/J\: /\: Km^2/W}$",fontsize=fontsize, fontweight="bold")
        else:
          plt.ylabel("$\mathbf{\Delta \overline{T}'}$ / K",fontsize=fontsize, fontweight="bold")
      plt.tick_params(axis='both', which='major', labelsize=fontsize-3)
      if elementvis:
        plt.grid()
    #plt.legend(fontsize=fontsize-3)
    if not nolabel:
      plt.xlabel("$\mathbf{\Delta z}$ / $\mathbf{\AA}$",fontsize=fontsize, fontweight="bold")
      #plt.xlabel("$\mathbf{\Delta z/r_b^{ref}}$", fontweight="bold", fontsize=fontsize) # TODP REMOVE THIS
    plt.tight_layout()
    plt.savefig(figname,dpi=200)
    outdic = {}
    outdic["allvals"] = allvals
    outdic["allvalst"] = allvalst
    return outdic
  
  # more standardized element plot function providing options
  # xvals     ... x coordinates of atoms
  # yvals     ... y coordinates of atoms
  # types     ... element types of atoms
  # othercol  ... color to plot elements with no default color
  # noelcol   ... do not color these with default colors
  def visualize_elems(self,xvals,yvals,types,othercol="k",noelcol = [],labels=False,xoffset=0,yoffset=0,alpha=1):
    a0 = 1.0
    alltypes = np.unique(types)
    for t in alltypes:
      logical = (types == t)
      if (t in self.defaultcols.keys()) & (t not in noelcol):
        pcol = self.defaultcols[t]
      else:
        pcol = othercol
      if labels:
        lab = t
      else:
        lab = ""
      plt.plot((xvals[logical]+xoffset)/a0,yvals[logical]+yoffset,"o",color=pcol,label=lab,markeredgecolor="k",markeredgewidth=1,alpha=alpha,markersize=10)
    return
  
  
  # set the heat flux at a certain distance from end
  def set_flux(self,n_convtest=100,useflux=90):
    # obtain flux at certain distance from end (90 % has shown nice convergence while not being too early for a 5 ns simulation)
    fluxdists = np.asarray(list(map(int,np.floor(np.linspace(0,len(self.fluxt)-1,n_convtest)))))
    favg = np.zeros(np.shape(fluxdists))
    for i,fdist in enumerate(fluxdists):
      favg[i] = np.mean(self.fluxt[fdist:])
    self.usedflux = favg[useflux]
    return favg
  
  # method to obtain the thermal resistances from step profile evaluation (step overlay)
  # assumes only two segments for now
  def obtain_thermal_resistances(self,midseg,n_convtest=100, useflux=90):
    
    favg = self.set_flux(n_convtest,useflux)
    
    midid = np.where(self.segnames==midseg)[0]
    midname = self.segnames[midid]
    if (len(self.segnames) < max(midid+1)):
      othername = self.segnames[midid+1]
    else:
      othername = self.segnames[midid-1]
    
    midkappa = favg[useflux]/self.allfits[1][0]*self.units_grad
    otherkappa = favg[useflux]/self.allfits[0][0]*self.units_grad
    interfacekapitza = favg[useflux]/np.mean(self.Tdiffs)
    self.midres       = self.units_grad/midkappa*self.fitseglens[1]
    self.otherres     = self.units_grad/otherkappa*self.fitseglens[0]
    self.interfaceres = 1.0/interfacekapitza
    self.midreserr       = self.fiterrs[1]/favg[useflux]*self.fitseglens[1]
    self.otherreserr     = self.fiterrs[0]/favg[useflux]*self.fitseglens[0]
    self.interfacereserr = np.mean(self.Tdifferrs)/favg[useflux]
    self.usedflux = favg[useflux]
    print (self.midreserr,self.otherreserr,self.interfacereserr)
    #print (self.midres, self.otherres, self.interfaceres)
    
    
    return
  
    
  
  # obtain thermal resistances for a system with only nodes
  def obtain_ln_thermal_resistances(self,midseg,n_convtest=100, useflux=90):
    
    favg = self.set_flux(n_convtest,useflux)
    
    midid = np.where(self.segnames==midseg)[0]
    midname = self.segnames[midid]
    if (len(self.segnames) < max(midid+1)):
      othername = self.segnames[midid+1]
    else:
      othername = self.segnames[midid-1]
    
    midkappa = favg[useflux]/self.allfits[1][0]*self.units_grad
    otherkappa = favg[useflux]/self.allfits[0][0]*self.units_grad
    self.midres       = self.units_grad/midkappa*self.fitseglens[1]
    self.otherres     = self.units_grad/otherkappa*self.fitseglens[0]
    self.interfaceres = None
    self.midreserr       = self.fiterrs[1]/favg[useflux]*self.fitseglens[1]
    self.otherreserr     = self.fiterrs[0]/favg[useflux]*self.fitseglens[0]
    self.interfacereserr = None
    self.usedflux = favg[useflux]
    print (self.midreserr,self.otherreserr,self.interfacereserr)
    #print (self.midres, self.otherres, self.interfaceres)
    
    
    return
  
  # method to visualize thermal resistance values in the shape of a bar plot
  # need execution of obtain_thermal_resistances first
  def plot_res_bars(self,offset=0,legend_labels=["Node","2xinterface","Linker"],horizontal=True,capsize=5,spacing=1,intcol="C3"):
    node = self.midres/1e-6
    linker = self.otherres/1e-6
    if self.interfaceres != None:
      interf = self.interfaceres*2/1e-6
    # this is how to do it in case we want to plot bars on top of each other
    #plt.bar(bars,node,label="Node")
    #plt.bar(bars,linker,bottom=node,label="Linker")
    #plt.bar(bars,interf,bottom=node+linker,label="2xInterface")
    if horizontal:
      pltfunc = plt.barh
      pltfunc((offset+0.0)*spacing,node,  0.2*spacing,label=legend_labels[0],capsize=capsize,xerr=self.midreserr/1e-6,        color=self.allfitcol[1])
      if self.interfaceres != None:
        pltfunc((offset+0.2)*spacing,interf,0.2*spacing,label=legend_labels[1],capsize=capsize,xerr=self.interfacereserr*2/1e-6,      color=intcol)
      pltfunc((offset+0.4)*spacing,linker,0.2*spacing,label=legend_labels[2],capsize=capsize,xerr=self.otherreserr/1e-6,color=self.allfitcol[0])
    else:
      pltfunc = plt.bar
      pltfunc(spacing*(offset+0.0),node,  0.2*spacing,label=legend_labels[0],capsize=capsize,yerr=self.midreserr/1e-6,        color=self.allfitcol[1])
      if self.interfaceres != None:
        pltfunc(spacing*(offset+0.2),interf,0.2*spacing,label=legend_labels[1],capsize=capsize,yerr=self.interfacereserr*2/1e-6,      color=intcol)
      pltfunc(spacing*(offset+0.4),linker,0.2*spacing,label=legend_labels[2],capsize=capsize,yerr=self.otherreserr/1e-6,color=self.allfitcol[0])
    return
  
  
  
      
